"""
把所有client模块以外从外部引用的类定义引用到这里，方便把haiddf拷贝到hepai项目中使用
"""


# from ...controller.worker_manager._worker_class import (
#     ModelResourceInfo, WorkerStatusInfo, WorkerNetworkInfo, WorkerInfo,
#     HWorkerModel, WorkerStoppedInfo,
# )

# from ...controller.api_gateway._request_class import WorkerInfoRequest, WorkerUnifiedGateRequest

from ..base_class._worker_class import (
    ModelResourceInfo, WorkerStatusInfo, WorkerNetworkInfo, WorkerInfo,
    HWorkerModel, WorkerStoppedInfo
)

from ..base_class._request_class import WorkerInfoRequest, WorkerUnifiedGateRequest





from typing import Literal, List, Dict, Union
from ._base_resource import SyncAPIResource
from .._client_class import HAPIKeyListPage
from .._related_class import APIKeyInfo, APIKEYDeletedInfo

class KeyManager(SyncAPIResource):

    @property
    def path(self):
        return '/key'

    def info(self, api_key: str):
        """
        curl 'http://0.0.0.0:4000/key/info?key=<user-key>' \
        -X GET \
        -H 'Authorization: Bearer <your-master-key>
        """
        url = f'{self.base_url}{self.path}/info?key={api_key}'
        headers = self.headers
        return self._get(url, headers=headers)

    def list_api_keys(self):
        """
        列出用户所有的api-key
        """
        url = f'{self.base_url}{self.path}/list_api_keys'
        return self._get(url, headers=self.headers, cast_to=HAPIKeyListPage)
    
    def create_api_key(
            self,
            *,
            key_name: str = "Default",
            valid_time: int = 30,
            user_id: str = None,
            allowed_models: Union[str, Dict] = "all",
            remarks: str = "",
            ):
        """
        创建API Key
        """
        url = f'{self.base_url}{self.path}/create_api_key'
        payload = {
            "key_name": key_name,
            "valid_time": valid_time,
            "user_id": user_id,
            "allowed_models": allowed_models,
            "remarks": remarks,
        }
        return self._post(
            url,
            headers=self.headers,
            json=payload,
            cast_to=APIKeyInfo,
            )
    
    def delete_api_key(self, api_key_id: str):
        """
        删除API Key
        """
        url = f'{self.base_url}{self.path}/delete_api_key'
        payload = {
            "api_key_id": api_key_id,
        }
        return self._post(
            url,
            headers=self.headers,
            json=payload,
            cast_to=APIKEYDeletedInfo,
            )

    ### --- 以下的LiteLLM的，准备弃用 --- ###
    def generate(
            self,
            *,
            duration: Literal["30s", "30m", "30h", "30d"] = None,
            key_alias: str = None,
            key: str = None,
            team_id: str = None,
            user_id: str = None,
            models: List[str] = [],
            aliases: Dict = {},
            config: Dict = {},
            spend: int = 0,
            send_invite_email: bool = False,
            max_budget: float = None,
            budget_duration: Literal["30s", "30m", "30h", "30d"] = None,
            max_parallel_requests: int = None,
            metadata: Dict = {},
            guardrails: List[str] = None,
            permissions: Dict = {},
            model_max_budget: Dict = {},
            model_rpm_limit: Dict = {},
            model_tpm_limit: Dict = {},
            ):
        """
        Generate an API key based on the provided data.
        Args:
            duration: Optional[str] - Specify the length of time the token is valid for. You can set duration as seconds ("30s"), minutes ("30m"), hours ("30h"), days ("30d").
            key_alias: Optional[str] - User defined key alias
            key: Optional[str] - User defined key value. If not set, a 16-digit unique sk-key is created for you.
            team_id: Optional[str] - The team id of the key
            user_id: Optional[str] - The user id of the key
            models: Optional[list] - Model_name's a user is allowed to call. (if empty, key is allowed to call all models)
            aliases: Optional[dict] - Any alias mappings, on top of anything in the config.yaml model list. - https://docs.litellm.ai/docs/proxy/virtual_keys#managing-auth---upgradedowngrade-models
            config: Optional[dict] - any key-specific configs, overrides config in config.yaml
            spend: Optional[int] - Amount spent by key. Default is 0. Will be updated by proxy whenever key is used. https://docs.litellm.ai/docs/proxy/virtual_keys#managing-auth---tracking-spend
            send_invite_email: Optional[bool] - Whether to send an invite email to the user_id, with the generate key
            max_budget: Optional[float] - Specify max budget for a given key.
            budget_duration: Optional[str] - Budget is reset at the end of specified duration. If not set, budget is never reset. You can set duration as seconds ("30s"), minutes ("30m"), hours ("30h"), days ("30d").
            max_parallel_requests: Optional[int] - Rate limit a user based on the number of parallel requests. Raises 429 error, if user's parallel requests > x.
            metadata: Optional[dict] - Metadata for key, store information for key. Example metadata = {"team": "core-infra", "app": "app2", "email": "ishaan@berri.ai" }
            guardrails: Optional[List[str]] - List of active guardrails for the key
            permissions: Optional[dict] - key-specific permissions. Currently just used for turning off pii masking (if connected). Example - {"pii": false}
            model_max_budget: Optional[dict] - key-specific model budget in USD. Example - {"text-davinci-002": 0.5, "gpt-3.5-turbo": 0.5}. IF null or {} then no model specific budget.
            model_rpm_limit: Optional[dict] - key-specific model rpm limit. Example - {"text-davinci-002": 1000, "gpt-3.5-turbo": 1000}. IF null or {} then no model specific rpm limit.
            model_tpm_limit: Optional[dict] - key-specific model tpm limit. Example - {"text-davinci-002": 1000, "gpt-3.5-turbo": 1000}. IF null or {} then no model specific tpm limit. Examples:
        """
        # 定义请求的 URL 和头部信息
        url = f'{self.base_url}/key/generate'
        headers = self.headers
   
        payload = {
            "duration": duration,
            "key_alias": key_alias,
            "key": key,
            "team_id": team_id,
            "user_id": user_id,
            "models": models,
            "aliases": aliases,
            "config": config,
            "spend": spend,
            "send_invite_email": send_invite_email,
            "max_budget": max_budget,
            "budget_duration": budget_duration,
            "max_parallel_requests": max_parallel_requests,
            "metadata": metadata,
            "guardrails": guardrails,
            "permissions": permissions,
            "model_max_budget": model_max_budget,
            "model_rpm_limit": model_rpm_limit,
            "model_tpm_limit": model_tpm_limit,
        }

            
        return self._post(
            url, 
            headers=headers,
            json=payload,
            )
    
    



from typing import Any, Dict, List, Tuple, Union
from dataclasses import dataclass
from io import BytesIO
from PIL import Image
import numpy as np

@dataclass
class HaiFile:
    """
    a file object for parsing the file content
    """
    type_: str
    data: Any
    filename: str = None

    def __post_init__(self):
        if self.type_ == 'image':
            self.data = BytesIO(self.data)
            img = Image.open(self.data)
            self.data = np.array(img)
        # elif self.type_ == 'txt':
        #     self.data = self.data.read().decode('utf-8')

    @property
    def type(self):
        return self.type_

    def read(self):
        return self.data
    
    def save(self, file_path: str=None, save_dir: str = None, debug=False):
        if file_path is None:
            if save_dir is None:
                file_path = self.filename
            else:
                file_path = f'{save_dir}/{self.filename}'
        if not file_path:
            raise ValueError("Please specify the file path, or the save_dir and filename.")
        if self.type_ == 'image':
            img = Image.fromarray(self.data)
            img.save(file_path)
        else:
            with open(file_path, 'wb') as f:
                f.write(self.data)
        if debug:
            print(f"Save file to {file_path}")

    







import os, sys
import hai

def request_model(prompt='hello', system_prompt=None):
    system_prompt = system_prompt if system_prompt else "Answering questions conversationally"

    result = hai.LLM.chat(
            # model='openai/gpt-3.5-turbo',
            model='hepai/vicuna-13B',
            messages=[
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": prompt},
                ## 如果有多轮对话，可以继续添加，"role": "assistant", "content": "Hello there! How may I assist you today?"
                ## 如果有多轮对话，可以继续添加，"role": "user", "content": "I want to buy a car."
            ],
            stream=True,
        )

    full_result = ""
    for i in result:
        full_result += i
        sys.stdout.write(i)
        sys.stdout.flush()
    print()
    return full_result

prompt = 'who are you?'
prompt = 'hello'
answer = request_model(prompt=prompt)



import os
from typing import Any, Dict, Union, List
from dataclasses import dataclass

from functools import wraps
import httpx
from .resources._base_resource import HErrorResponse

def parse_data_to_class_deprecated(data: Dict[str, Any], target_class: Any):
    """
    Parses data into an instance of the target class.
    
    :param data: Dictionary containing the data.
    :param target_class: The class to which the data is to be parsed.
    :return: An instance of target_class initialized with the data.
    """
    # Filter the data dictionary to only include keys that are in the target class
    filtered_data = {
        key: value for key, value in data.items() if key in target_class.__annotations__
    }
    
    # Handle nested dataclasses if needed
    for field_name, field_type in target_class.__annotations__.items():
        # if isinstance(filtered_data.get(field_name), List):
        #     ## List内可能还有Dict
        #     inner_type = getattr(field_type, '__args__', [Any])[0]

        #     pass
        if isinstance(filtered_data.get(field_name), dict) and hasattr(field_type, '__dataclass_fields__'):
            filtered_data[field_name] = parse_data_to_class_deprecated(filtered_data[field_name], field_type)

    # Creating an instance of the target class with the filtered data
    return target_class(**filtered_data)

def parse_data_to_class(data: Dict[str, Any], target_class: Any):
    """
    Parses data into an instance of the target class.
    
    :param data: Dictionary containing the data.
    :param target_class: The class to which the data is to be parsed.
    :return: An instance of target_class initialized
    with the data.
    """
    return target_class(**data)

def post_process_decorator(func) -> Union[str, Dict, HErrorResponse]:
    """
    Dealing with the response of http request
    
    Args:
        cast_to: The class to which the data is to be parsed.
    """
    dev_mode = os.getenv("DEV_MODE", False)

    @wraps(func)
    def wrapper(*args, **kwargs):
        cast_to = kwargs.pop("cast_to", None)
        ignore_error = kwargs.pop("ignore_error", False)

        response: httpx.Response = func(*args, **kwargs)
        if response.status_code == 200:
            try:
                rst = response.json()
            except ValueError:
                # If JSON decoding fails, return the text content
                return response.text
            if cast_to is None:
                return rst
            else:
                # 将Json数据转换为指定的数据类
                return parse_data_to_class(rst, cast_to)
        elif response.status_code == 500:
            raise HErrorResponse(
                status_code=response.status_code,
                message=response.reason_phrase or "Unknown error",
                content=response.text
            )
        else:

            if dev_mode and not ignore_error:  # 如果debug, 就在这里直接报错
                # Construct an HErrorResponse and raise it
                raise HErrorResponse(
                    status_code=response.status_code,
                    message=response.reason_phrase or "Unknown error",
                    content=response.text
                )
            # Construct an HErrorResponse and return it
            return HErrorResponse(
                status_code=response.status_code,
                message=response.reason_phrase or "Unknown error",
                content=response.text
            )
    return wrapper



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

try:
        from . import grpc_pb2 as grpc__pb2
except:
        import grpc_pb2 as grpc__pb2



class GrpcServiceStub(object):
    """定义服务接口
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.hello = channel.unary_unary(
                '/GrpcService/hello',
                request_serializer=grpc__pb2.HelloRequest.SerializeToString,
                response_deserializer=grpc__pb2.HelloResponse.FromString,
                )
        self.call = channel.unary_unary(
                '/GrpcService/call',
                request_serializer=grpc__pb2.CallRequest.SerializeToString,
                response_deserializer=grpc__pb2.CallResponse.FromString,
                )


class GrpcServiceServicer(object):
    """定义服务接口
    """

    def hello(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def call(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GrpcServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'hello': grpc.unary_unary_rpc_method_handler(
                    servicer.hello,
                    request_deserializer=grpc__pb2.HelloRequest.FromString,
                    response_serializer=grpc__pb2.HelloResponse.SerializeToString,
            ),
            'call': grpc.unary_unary_rpc_method_handler(
                    servicer.call,
                    request_deserializer=grpc__pb2.CallRequest.FromString,
                    response_serializer=grpc__pb2.CallResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'GrpcService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class GrpcService(object):
    """定义服务接口
    """

    @staticmethod
    def hello(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GrpcService/hello',
            grpc__pb2.HelloRequest.SerializeToString,
            grpc__pb2.HelloResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def call(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GrpcService/call',
            grpc__pb2.CallRequest.SerializeToString,
            grpc__pb2.CallResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)







from typing import Literal, List, Dict, Any
import types
import functools


from .._related_class import (
    WorkerInfo, HWorkerModel, WorkerStoppedInfo, WorkerStatusInfo,
)
from ._base_resource import SyncAPIResource, HErrorResponse
from .._client_class import HWorkerListPage


class WorkerResource(SyncAPIResource):

    @property
    def path(self) -> str:
        """path of the worker resource, i.e. the prefix of the endpoint"""
        return "/worker"

    def info(self, worker_id: str=None, model_name: str=None, **kwargs) -> WorkerInfo:
        url = f'{self.base_url}{self.path}/get_worker_info'
        assert worker_id or model_name, "Either worker_id or model_name should be provided."
        payload = {
            "worker_id": worker_id,
            "model_name": model_name,
        }
        return self._post(
            url,
            json=payload,  # 写成params而非json是因为后端接收的是query string
            cast_to=WorkerInfo,
            **kwargs,
        )
    
    def status(self, worker_id: str, refresh: bool = False) -> WorkerStatusInfo:
        url = f'{self.base_url}{self.path}/get_worker_status'
        payload = {
            "worker_id": worker_id,
            "refresh": refresh,
        }
        return self._post(
            url,
            json=payload,
            cast_to=WorkerStatusInfo,
        )
    
    def request(
            self,
            # target: dict,
            params: dict,
            args: list,
            kwargs: dict,
            ):
        url = f'{self.base_url}{self.path}/unified_gate'
        payload = {
            # "target": target,
            "args": args,
            "kwargs": kwargs,
        }
        return self._post(
            url,
            params=params,
            json=payload,
        )
    
    def list(self, url: str=None) -> HWorkerListPage:
        url = f'{self.base_url}{self.path}/list_workers' if url is None else f'{self.base_url}{url}'
        return self._get(
            url,
            cast_to=HWorkerListPage,
            )
    
    def refresh_all(self) -> Dict:
        url = f'{self.base_url}{self.path}/refresh_all_workers'
        return self._get(
            url,
            cast_to=dict,
            )
    
    def register(
            self, 
            model: Dict, 
            daemon: bool=False,
            standalone: bool=False,
        ) -> WorkerInfo:
        
        from ...worker.worker_app import HWorkerAPP
        return HWorkerAPP.register_worker(model=model, daemon=daemon, standalone=standalone)
    
    
    def stop(self, worker_id: str, permanent: bool=False) -> WorkerStoppedInfo:
        url = f'{self.base_url}{self.path}/stop_worker'
        payload = {
            "worker_id": worker_id,
            "permanent": permanent,
        }
        return self._post(
            url,
            json=payload,
            cast_to=WorkerStoppedInfo,
            )

    def get_remote_model(self, model_name: str) -> "HRemoteModel":
        """
        Get remote model by model name
        """
        return HRemoteModel(model_name=model_name, worker_resource=self)
        

class HRemoteModel:

    def __init__(
            self,
            model_name: str,
            worker_resource: WorkerResource,
            ) -> None:
        self.model_name = model_name
        self.wr: WorkerResource = worker_resource
        worker_info: WorkerInfo = self.wr.info(model_name=model_name, worker_id=None, ignore_error=True)
        if isinstance(worker_info, HErrorResponse):
            raise ValueError(f"Failed to get remote model: {worker_info}")

        model_functions = worker_info.resource_info.model_functions
        # model_functions = ["train", "__call__"]
        if len(model_functions) == 0:
            raise ValueError(f"Remote model `{model_name}` has no functions can be called remotely, please check the worker model.")

        # 自动注册远程模型的函数
        for func in model_functions:
            original_func = self.function_warpper(model_name, func)
            # Use functools.wraps to preserve original function metadata when creating new methods
            # @functools.wraps(original_func)
            # def wrapper(*args, **kwargs):
            #     return original_func(*args, **kwargs)
            # setattr(self, func, types.MethodType(wrapper, self))
            setattr(self, func, types.MethodType(original_func, self))
   

    def function_warpper(self, model_name: str, function_name: str):
        def call_remote_function(*args, **kwargs):
            if isinstance(args[0], HRemoteModel):
                # 为了处理通过types.MethodType注册时，第一个参数是self的情况
                args = args[1:]
            return self.wr.request(
                params={
                    "model": model_name, 
                    "function": function_name
                    },
                args=args,
                kwargs=kwargs,
            )
        return call_remote_function
    
    def __call__(self, *args, **kwargs):
        return self.function_warpper(self.model_name, '__call__')(*args, **kwargs)

  


"""
调用Worker的API
"""


import os, sys
from pathlib import Path
import requests
import hai


class HaiLLM(object):

    @staticmethod
    def chat(model, messages=None, **kwargs):
        """Create a LLM instance.

        :param model: The model name.
        :param messages: The messages.
        :return: The LLM instance.
        """
        api_key = kwargs.pop("api_key", None)
        api_key = api_key or hai.api_key

        url = kwargs.pop("url", None)
        if not url:
            host = kwargs.pop("host", "aiapi.ihep.ac.cn")
            port = kwargs.pop("port", None)
            if port is not None:
                url = f"http://{host}:{port}"
            else:
                url = f"https://{host}"
        
        
        data = dict()
        data['model'] = model
        data['messages'] = messages
        data['stream'] = kwargs.pop('stream', True)
        data.update(kwargs)

        assert api_key, """
The HepAI API-KEY is required. Please set the environment variable `HEPAI_API_KEY` via `export HEPAI_API_KEY=xxx`.
Alternatively, it can be provided by passing in the `api_key` parameter when calling the `chat` method.
"""
        session = requests.Session()
        response = session.post(
            # f'{url}/v1/chat/completions',
            f'{url}/v1/inference',
            headers={"Authorization": f"Bearer {api_key}"},
            json=data,
            stream=True,
            timeout=60,
            )
        # print(f'llm response: {response}')
        full_response = ""
        # print('llm streaming:')
        for chunk in response.iter_lines(decode_unicode=False, delimiter=b"\0"):
            if not chunk:
                continue
            chunk = chunk.decode('utf-8')
            if chunk == "[DONE]":
                break
            full_response += chunk
            # print(f'\r{full_response}', end='')
            yield chunk

        # print('\n')
        # print(f'full_response: {full_response}')
        # print(model, messages)


if __name__ == '__main__':
    import sys
    model = 'hepai/chathep-0527'
    model = 'openai/gpt-3.5-turbo'
    api_key = os.getenv('HEPAI_API_KEY')
    messages = [
        {'role': 'system', 'content': 'You are a bot.'},
        {'role': 'user', 'content': 'Hello, how are you?'}
    ]
    # messages = [
    #     {"role": "system", "content": 'Yo'},
    #     {"role": "user", "content": 'Hello'},
    #     ## 如果有多轮对话，可以继续添加，"role": "assistant", "content": "Hello there! How may I assist you today?"
    #     ## 如果有多轮对话，可以继续添加，"role": "user", "content": "I want to buy a car."
    # ]
    result =  HaiLLM.chat(
        model,
        api_key=api_key, 
        messages=messages,
        host="aiapi.ihep.ac.cn", port=42901,
        )
    full_result = ""
    for i in result:
        full_result += i
        sys.stdout.write(i)
        sys.stdout.flush()
    print()



try:
    from .hf_argparser import HfArgumentParser
except:
    from hf_argparser import HfArgumentParser
    
def parse_args(dataclasses_types,
        args=None,
        return_remaining_strings=False,
        look_for_args_file=True,
        args_filename=None,
        args_file_flag=None,):
    """
    Parse command-line args into instances of the specified dataclass types.

    This relies on argparse's `ArgumentParser.parse_known_args`. See the doc at:
    docs.python.org/3.7/library/argparse.html#argparse.ArgumentParser.parse_args

    Usage::
        python xx.py --arg1 val1 --arg2 val2 

    Args:
        dataclass_types:
            Dataclass type, or list of dataclass types for which we will "fill" instances with the parsed args.
        args:
            List of strings to parse. The default is taken from sys.argv. (same as argparse.ArgumentParser)
        return_remaining_strings:
            If true, also return a list of remaining argument strings.
        look_for_args_file:
            If true, will look for a ".args" file with the same base name as the entry point script for this
            process, and will append its potential content to the command line args.
        args_filename:
            If not None, will uses this file instead of the ".args" file specified in the previous argument.
        args_file_flag:
            If not None, will look for a file in the command-line args specified with this flag. The flag can be
            specified multiple times and precedence is determined by the order (last one wins).

    Returns:
        Tuple consisting of:

            - the dataclass instances in the same order as they were passed to the initializer.abspath
            - if applicable, an additional namespace for more (non-dataclass backed) arguments added to the parser
                after initialization.
            - The potential list of remaining argument strings. (same as argparse.ArgumentParser.parse_known_args)
    """
    return parse_args_into_dataclasses(
        dataclasses_types,
        args=args,
        return_remaining_strings=return_remaining_strings,
        look_for_args_file=look_for_args_file,
        args_filename=args_filename,
        args_file_flag=args_file_flag,
    )
    

def parse_args_into_dataclasses(
        dataclasses_types,
        args=None,
        return_remaining_strings=False,
        look_for_args_file=True,
        args_filename=None,
        args_file_flag=None,):
    """
    Parse command-line args into instances of the specified dataclass types.

    This relies on argparse's `ArgumentParser.parse_known_args`. See the doc at:
    docs.python.org/3.7/library/argparse.html#argparse.ArgumentParser.parse_args

    Usage::
        python xx.py --arg1 val1 --arg2 val2 

    Args:
        dataclass_types:
            Dataclass type, or list of dataclass types for which we will "fill" instances with the parsed args.
        args:
            List of strings to parse. The default is taken from sys.argv. (same as argparse.ArgumentParser)
        return_remaining_strings:
            If true, also return a list of remaining argument strings.
        look_for_args_file:
            If true, will look for a ".args" file with the same base name as the entry point script for this
            process, and will append its potential content to the command line args.
        args_filename:
            If not None, will uses this file instead of the ".args" file specified in the previous argument.
        args_file_flag:
            If not None, will look for a file in the command-line args specified with this flag. The flag can be
            specified multiple times and precedence is determined by the order (last one wins).

    Returns:
        Tuple consisting of:

            - the dataclass instances in the same order as they were passed to the initializer.abspath
            - if applicable, an additional namespace for more (non-dataclass backed) arguments added to the parser
                after initialization.
            - The potential list of remaining argument strings. (same as argparse.ArgumentParser.parse_known_args)
    """
    # assert isinstance(dataclasses_types, tuple), f'input dataclasses must be tuple, but got {type(dataclasses_types)}'
    is_tuple = isinstance(dataclasses_types, tuple)
    if not is_tuple:
        dataclasses_types = (dataclasses_types, )
    result = HfArgumentParser(dataclasses_types).parse_args_into_dataclasses(
        args=args,
        return_remaining_strings=return_remaining_strings,
        look_for_args_file=look_for_args_file,
        args_filename=args_filename,
        args_file_flag=args_file_flag,
    )
    if is_tuple:
        return result
    else:
        r1, = result
        return r1


def test():
        
    import dataclasses
    @dataclasses.dataclass
    class TestArgs:
        name: str = 'zhangsan'
        age: int = 18

    @dataclasses.dataclass
    class TestArgs2:
        name2: str = 'lisi'

    args = parse_args_into_dataclasses(TestArgs)
    print(args)
    print(args.name)

    args1, args2 = parse_args_into_dataclasses((TestArgs, TestArgs2))
    print(args1)
    print(args2)
    print(args2.name2)


if __name__ == '__main__':
    test()








from .grpc import grpc_secure_server
import argparse
import damei as dm

logger = dm.get_logger('server__init__')


def get_is_running(port):
    ret = dm.popen(f'netstat -tanlp')
    ret = [x.split() for x in ret if len(x.split()) == 7]
    # port = str(opt.port)
    ports = [f'{x[3].split(":")[-1]}' for x in ret]
    if port in ports:  # 有正在运行的服务
        index = ports.index(port)
        pid = ret[index][-1]
        if 'python' not in pid:  # 如果不是python进程，则不是xai服务
            logger.warn(f'port {port} is running, but not xai-server')
            return False
        else:
            pid = pid.split('/')[0]
            # dm.popen(f'kill -9 {pid}')
            return pid
    else:
        # logger.error(f'xai-server on port {port} is not running')
        return False


def run(opt):
    # grpc_xai_server.run(ip=ip, port=port)
    # logger.info(f'xai version: {hai.__version__}')
    logger.info(f'Xai-server params: {opt}')
    assert opt.mode in ['start', 'stop', 'restart'], 'mode must be start, stop or restart'
    if opt.mode == 'start':
        if opt.insecure:
            grpc_secure_server.run_insecure(port=opt.port, debug=opt.debug)
        else:
            grpc_secure_server.run(port=opt.port, debug=opt.debug)
    elif opt.mode == 'stop':
        is_running = get_is_running(str(opt.port))
        # print(is_running)
        if is_running:  # 如果有运行的服务，返回的是pid
            dm.popen(f'kill -9 {is_running}')
        else:
            logger.warn(f'xai-server on port {opt.port} is not running')
            sys.exit(0)
    else:
        is_running = get_is_running(str(opt.port))
        if is_running:
            dm.popen(f'kill -9 {is_running}')
        if opt.insecure:
            grpc_secure_server.run_insecure(port=opt.port)
        else:
            grpc_secure_server.run(port=opt.port)


import os, sys
from pathlib import Path
from easydict import EasyDict

import functools
import collections
import copy
import re
import damei as dm
import hai
import json
import argparse

pydir = Path(os.path.abspath(__file__)).parent
logger = dm.get_logger(__name__)


class PyConfigLoader(collections.UserDict):
# class PyConfigLoader(object):
    """
    从字典、.py文件等获取的配置对象
    v1.0.1 支持元数据
    """

    def __init__(self, cfg_file=None, name=None, root_path=None, cfg_dict=None, **kwargs):
        super(PyConfigLoader, self).__init__()
        # super().__init__()
        work_dir = os.getcwd()
        self.root_path = root_path if root_path else f'{work_dir}'  # 默认当前工作目录
        self._name = name if name else f'{cfg_file}'  # 其实是path

        self._items = dict()
        self._load_items(cfg_file, cfg_dict, **kwargs)  # 从配置文件或配置字典初始化配置，配置添加到_items和对应的属性里
        self.check_items()  # TODO

    def _load_items(self, cfg_file=None, cfg_dict=None, **kwargs):
        if cfg_file is None and cfg_dict is None:
            return

        assert cfg_file or cfg_dict, 'Cfg_file or cfg_dict must be not None'
        assert not (cfg_file and cfg_dict), 'Only one of cfg_file and cfg_dict can have value'
        if cfg_file:
            assert os.path.exists(cfg_file), f'Config file {cfg_file} does not exist'
            self.init_config_by_file(cfg_file)  # 根据配置文件把内容和属性注册到items里
        else:
            self.init_config_by_dict(cfg_dict, **kwargs)

    def init_config_by_dict(self, cfg_file, **kwargs):
        """从字典初始化配置"""
        need_replace = kwargs.get('need_replace', False)  # 是否需要替换 - to _，默认不需要, needed when read config from dm.argparse shi xuyao
        for k, v in cfg_file.items():
            attr, attr_value = k, v
            if need_replace and '-' in attr:
                attr = attr.replace('-', '_')

            self._items[attr] = attr_value
            if isinstance(attr_value, dict):
                setattr(self, attr, EasyDict(self._items[attr]))
            else:
                setattr(self, attr, self._items[attr])

    def __len__(self):
        return len(self.items)

    # def __dict__(self):
    #   return self._items

    def __getitem__(self, item):
        return self.items[item]

    def __repr__(self):
        info = self.info(need_head=False, need_tail=False)
        if info == 'Empty config':
            format_str = f"Empty <class 'hai.Config'> "
        else:
            format_str = f"<class 'hai.Config'> " \
                        f"Configs:\n" \
                        f"{self.info(need_head=False, need_tail=False)}"

        return format_str

    def __missing__(self, key):
        if isinstance(key, str):
            raise KeyError(key)
        return self[str(key)]

    def __contains__(self, key):
        return str(key) in self.data

    def __setitem__(self, key, item):
        # print(f'__setitem__:{key}')
        # self.data[str(key)] = item
        self._items[str(key)] = item
        self.__setattr__(str(key), item)

    @property
    def path(self):
        return self._name

    @property
    def items(self):
        """
        debug: 
            when call self.some_attr = new_value, the new_value will be set to self.__dict__ only,
            but not to self._items.
            To fix this, we need to copy self.__dict__ to self._items every if the value is changed.
        """
        # for k, v in self._items.items():
        #     __dict__v = self.__dict__[k]
        #     if __dict__v != v:
        #         self._items[k] = __dict__v

        # return self._items
        return self.__dict__

    @staticmethod
    def from_anything(*args, **kwargs):
        """
        从字典、.py文件等获取的配置对象
        """
        if len(args) == 1:
            if isinstance(args[0], dict):
                return PyConfigLoader(cfg_dict=args[0], **kwargs)
            elif isinstance(args[0], str):
                return PyConfigLoader(cfg_file=args[0], **kwargs)
            else:
                raise NotImplementedError('Only support dict and str')
        return PyConfigLoader(**kwargs)

    @staticmethod
    def from_file(cfg_file, name=None, root_path=None):
        return PyConfigLoader(cfg_file, name, root_path)

    @staticmethod
    def fromfile(cfg_file, name=None, root_path=None):
        return PyConfigLoader(cfg_file, name, root_path)

    @staticmethod
    def from_dict(cfg_dict, name=None, root_path=None):
        """从字典创建配置对象"""
        name = name if name else 'cfg_from_dict'
        return PyConfigLoader(name=name, root_path=root_path, cfg_dict=cfg_dict)

    @staticmethod
    def from_argparser(args, name=None, root_path=None):
        """从hai.parser.ArgumentParser对象创建配置对象"""
        name = name if name else 'cfg_from_argparser'
        return PyConfigLoader(name=name, root_path=root_path, cfg_dict=args._args_dict)

    @staticmethod
    def from_argparse(opt, name=None, root_path=None):
        name = name if name else 'cfg_from_argparse'
        __dict__ = opt.__dict__
        # print(opt)
        return PyConfigLoader(name=name, root_path=root_path, cfg_dict=__dict__)

    @staticmethod
    def from_dm_argparse(opt, name=None, root_path=None):
        name = name if name else 'cfg_from_dm_argparse'
        __dict__ = opt._args_dict
        return PyConfigLoader(name=name, root_path=root_path, 
                            cfg_dict=__dict__, need_replace=True)


    def info(self, need_head=True, need_tail=True):
        """
        查看当前所有配置信息
        :param need_head: 是否需要包含头部信息，即：Configs:\n
        :param need_tail: 是否需要包含尾部信息，即：\n回车符
        """
        # info_str = f'{self.__repr__()}\n'
        info_dict = self._items
        if info_dict == {}:
            info_str = 'Empty config\n'
        else:
            info_str = f'Configs:\n' if need_head else ''
            info_str += dm.misc.dict2info(info_dict)
        if not need_tail:
            info_str = info_str[:-1]
        return info_str

    def init_config_by_file(self, cfg_file):
        if cfg_file is None:
            return
        rp = self.root_path
        cp = Path(os.path.abspath(cfg_file))  # config path

        if not os.path.exists(cfg_file):
            raise FileNotFoundError(f'{cfg_file} not found')

        # 提取模块路径：如：modules.detection.seyolov5.config
        method = 'method3'
        current_wd = os.getcwd()  # 当前工作目录
        if method == 'method1':
            module_dir = f'{str(cp.parent).replace(f"{rp}/", "").replace("/", ".")}'
            # print(f'rp: {rp} \ncp: {cp}\nm dir: {module_dir}')
            # code = f'from {module_dir} import {cp.stem}'
            code = f'import {module_dir}.{cp.stem} as {cp.stem}'
            # 不能用这种方法，原因是：from modules.xxx.config imort config 过程中会调起modules下的__init__.py,从而调起注册函数，导致重复注册
        elif method == 'method2':  # 方法2：添加临时环境路径
            # 这个方法在工作路径也存在config文件夹时，会导致from config import xxx 导入失败
            sys.path.append(f'{cp.parent.parent}')  # 添加临时环境路径, 如：/home/user/xx/modules/xx
            module_dir = f'{str(cp.parent.stem)}'  # configs文件夹路径，如：configs
            code = f'from {module_dir} import {cp.stem}'
        elif method == 'method3':  # 方法3：添加临时环境路径
            # sys.path.append(f'{cp.parent}')
            sys.path.insert(0, f'{cp.parent}')  # 添加临时环境路径, 如：/home/user/xx/modules/xx
            module_dir = '.'
            os.chdir(f'{cp.parent}')  # 切换工作目录
            # code = f'from {module_dir} import {cp.stem}'
            code = f'import {cp.stem}'
        else:
            raise ValueError(f'invalid method: {method}')

        # print(cfg_file, code)
        """
        exec(code)
        """
        try:
            exec(code)
        except Exception as e:
            msg = f'Error in "exec(code)": {e}. Error variables:\n' \
                  f'    root path: {rp} \n' \
                  f'    config file: {cfg_file} \n' \
                  f'    code: {code} \n' \
                  f'    module dir: {module_dir} \n' \
                  f'    sys.path[0]: {sys.path[0]} ' \
                  f'    sys.path[-1]: {sys.path[-1]}'
            # logger.error(msg)
            raise Exception(msg)
            # sys.exit(1)

        if method == 'method2':
            sys.path.remove(f'{cp.parent.parent}')  # 删除临时环境路径
        elif method == 'method3':
            sys.path.remove(f'{cp.parent}')
            os.chdir(current_wd)  # 切换回工作目录

        cfg = eval(cp.stem)  # 模块名

        # 读取文件获取顶格写并且不以#开头的属性
        with open(cfg_file, 'r', encoding='utf-8') as f:
            data = f.readlines()
        data = [x for x in data if not (x.startswith(' ') or x.startswith('#'))]  # 非顶格的不要，# 开头的不要
        data = '\n'+'\n'.join(data)
        attrs = re.findall(pattern=r"\s[A-Za-z]\w+ ?= ?", string=data)
        # The egular expression: \s[A-Za-z]\w+ ?= ? to match: ' attr = '
        # i.e. start with a space, then a letter (a to z or A to Z), 
        # then any number of letters, numbers or underscores, 
        # then a space, then an equal sign, then a space
        attrs = [x.split('=')[0].replace('\n', '').strip() for x in attrs]

        # 注册属性到_items里，注册属性到self的属性里。
        for attr in attrs:
            # print('xxx', attr)
            if hasattr(cfg, attr):
                exec(f"self._items[attr] = cfg.{attr}")
                attr_value = self._items[attr]
                if isinstance(attr_value, dict):
                    setattr(self, attr, EasyDict(self._items[attr]))
                else:
                    setattr(self, attr, attr_value)
            else:
                pass

    def check_items(self):
        """初始化末，检查所有配置项，处理：缩略语~转为真实路径。
        递归方法，如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环
        """
        # print(self._items)
        for k, v in self._items.items():
            new_v = self.recursive_func(v)
            self._items[k] = new_v
            # 更新属性
            if hasattr(self, k):
                delattr(self, k)
            setattr(self, k, EasyDict(new_v)) if isinstance(new_v, dict) else setattr(self, k, new_v)

    def recursive_func(self, value):
        """递归函数,如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环"""
        if value is None:
            return None
        elif isinstance(value, str):
            return value.replace('~', str(Path.home()))
        elif isinstance(value, int):
            return value
        elif isinstance(value, float):
            return value
        elif isinstance(value, bool):
            return value
        elif isinstance(value, list):
            new_list = [self.recursive_func(x) for x in value]
            return new_list
        elif isinstance(value, dict):
            new_dict = {}
            for k, v in value.items():
                new_dict[k] = self.recursive_func(v)
            return new_dict
        # elif isinstance(value, argparse.Namespace):  # TODO: suport argparse.Namespace
            # items = value.items()
        elif isinstance(value, hai.Config):
            new_dict = {}
            for k, v in value._items.items():
                new_dict[k] = self.recursive_func(v)
            new_config = PyConfigLoader.from_dict(new_dict)
            return new_config
        else:
            raise TypeError(f'Config file only support basic python types but the value type is: {type(value)}.')

    def update_item(self, attrs, value):
        """更新配置条目
        attrs: 长度为0时代表，更新全部value，长度为1时，代表第1个元素是第一层，
        """
        if len(attrs) == 0:
            self.merge(value)
        elif len(attrs) == 1:
            self._items[attrs[0]] = value
        elif len(attrs) == 2:
            self._items[attrs[0]][attrs[1]] = value
        elif len(attrs) == 3:
            self._items[attrs[0]][attrs[1]][attrs[2]] = value
        else:
            self._items[attrs[0]][attrs[1]][attrs[2]][attrs[3]] = value
        self.check_items()

    def merge2opt(self, opt):
        """
        合并配置文件到opt, 合并规则：没想好，先放着
        return: opt
        """
        assert isinstance(opt, argparse.Namespace), 'opt must be argparse.Namespace'
        raise NotImplementedError(f'未实现')

    def merge(self, cfg2, **kwargs):
        """合并另一个配置文件到内部，是inplace的"""
        if cfg2 is None:
            return self

        if isinstance(cfg2, dict):
            cfg2 = PyConfigLoader.from_dict(cfg2)

        for k2, v2 in cfg2._items.items():
            merged_v = self.recursive_func2(v2, k2=k2, sub_items=self._items)
            # self._items[k2] = merged_v
            self.__dict__[k2] = merged_v
        self.check_items()
        return self

    def recursive_func2(self, v2, k2=None, sub_items=None):
        """v1是之前的值，v2是现在的值"""
        if k2 and (k2 not in list(sub_items.keys())):
            return v2
        else:
            v1 = sub_items[k2]

            if type(v1) != type(v2):
                return v2
            f = functools.partial(isinstance, v1)
            if f(str) or f(int) or f(float) or f(bool):
                return v2
            elif f(list):
                tmp = [x for x in v2 if x not in v1]
                return v1 + tmp
            elif f(dict):  # v1v2都是dict
                new_dict = v1  # 保留v1
                for kk, vv in v2.items():
                    new_dict[kk] = self.recursive_func2(vv, k2=kk, sub_items=v1)
                return new_dict

    def map_keys(self, map_dict, new=True):
        """map_dict: {old_key: new_key}"""
        assert isinstance(map_dict, dict), 'map_dict must be dict'
        if new:
            new_cfg = copy.deepcopy(self)
            new_cfg.map_keys(map_dict, new=False)
            return new_cfg
        else:
            for k, v in map_dict.items():  # k是旧的，v是新的
                if k in self._items:
                    self.__dict__[v] = self.items[k]
                    del self.__dict__[k]
        

    def keys(self):
        return self._items.keys()

    def get(self, key, default=None):
        if key in list(self._items.keys()):
            return self._items[key]
        else:
            return default

    def to_dict(self):
        # raise NotImplementedError('to_dict is not implemented.')
        return self._items

    def to_json(self):
        return json.dumps(self._items, indent=4, ensure_ascii=False)



"""
保存流的信息
"""
import damei as dm
import numpy as np
from collections import OrderedDict
from .stream import Stream

logger = dm.getLogger('uaii streams')


class Streams(object):

    def __init__(self, cfg, parent):
        self.uaii_cfg = cfg  # 全部的配置
        self._cfg = self.uaii_cfg.streams
        self.parent = parent
        self._name = 'streams'
        self._module_dict = OrderedDict()
        self._init()

    def _init(self):
        cfg = self._cfg
        # print(cfg)
        for i, stream_cfg in enumerate(cfg):
            name = stream_cfg['type']
            # print(stream_cfg)
            stream = Stream(
                parent=self.parent, id=f'S{i + 1:0>2}', stream_cfg=stream_cfg)
            self._module_dict[name] = stream
            # print('xx', stream.seyolov5)

    @property
    def name(self):
        """self name: streams"""
        return self._name

    @property
    def cfg(self):
        """streams cfg"""
        return self._cfg

    @property
    def module_dict(self):
        """OrderedDict, key is the name, value is the stream object"""
        return self._module_dict

    @property
    def ids(self):
        ids = [x[0] for x in self.ps()]
        return ids

    @property
    def names(self):
        """module names"""
        names = [x[2] for x in self.ps()]
        return names

    def verify_input_cfg(self, cfg):
        """校验必要键：type, models"""
        keys = list(cfg.keys())
        if 'type' not in keys:
            raise ValueError(f'Config error. Stream cfg must specify "type", keys: {keys}')
        if 'models' not in keys:
            raise ValueError(f'Config error. Stream cfg must specify "models", keys: {keys}')

    def build_new_stream(self, cfgs, **kwargs):
        """
        build new stream by stream cfgs
        :param cfgs: list of module cfgs
            for example:
                [stream1, stream2, stream3], where streamx is a dict
        """
        # print(cfgs, type(cfgs))
        if isinstance(cfgs, dict):
            cfgs = [cfgs]

        # 获取最大的id
        ids = self.ids
        max_id = np.max([int(x[1::]) for x in self.ids]) if len(ids) != 0 else 0  # [s01, s02, s03, ...]
        # print(len(cfgs))
        # is_mono = len(cfgs) == 1
        for i, stream_cfg in enumerate(cfgs):  # 这里是单个或多个流
            # 校验stream_cfg的必须字段是否存在
            self.verify_input_cfg(stream_cfg)

            name = stream_cfg['type']
            is_mono = len(stream_cfg['models']) == 1
            stream = Stream(
                parent=self.parent,
                id=f'S{max_id + i + 1:0>2}',
                stream_cfg=stream_cfg,  # 这里是每个流有单个或多个模型
                is_mono=is_mono, )
            self._module_dict[name] = stream

        # 合并cfg，列表
        self._cfg += cfgs

    def get(self, module_name):
        return self._module_dict[module_name]

    def ps(self, module=None, **kwargs):
        """'ID', 'TYPE', 'NAME', 'STATUS', 'TAG', 'INCLUDE''DESCRIPTION'"""
        include_dict = kwargs.get('include_dict', None)
        modules = module if module else list(self.module_dict.keys())
        if isinstance(modules, str):
            modules = [modules]

        ret = []
        for module_names in modules:
            module = self.get(module_names)
            id = module.id
            type = 'stream'
            name = module.name
            status = module.status
            description = module.description

            include = module.include(format=list)
            if include_dict:  # 字典：模块名转为id
                include = [include_dict[x] for x in include]
                include = f"[{']['.join(include)}]"
                # print(include_dict, include)
            else:
                include = ' '.join(include)
            # config = module.cfg
            # ret.append([id, type, name, status, include, description, config])
            ret.append([id, type, name, status, '-', include, description])
        return ret



"""
默认的uaii流的配置
"""

# streams命名规则：键名需包含字符串stream, 特殊键名type
# modules命名规则：键名必须包含type，指定模块名字，列表前后顺序表明模块的前后顺序
stream1 = dict(
    type='visible_detection_stream',  # 流的名字
    description='可见光目标检测流',  # 流的描述
    models=[  # 流包含的模型及其配置文件
        dict(type='seyolov5',
             cfg='xsensing_ai/config/uaii_seyolov5_config.py'),  # 或字符串指定到模块的配置文件xxx.py
    ])

stream2 = dict(
    type='visible_tracking_stream',
    description='可见光目标检测+跟踪流',
    models=[  # 目标检测+跟踪
        dict(type='seyolov5',
             cfg='xsensing_ai/config/uaii_seyolov5_config.py'),
        dict(type='deepsort',
             cfg='xsensing_ai/config/uaii_deepsort_config.py')
    ])

stream3 = dict(
    type='infrared_detection_stream',
    description='红外目标检测流',
    models=[  # 红外目标检测
        dict(type='irdet',
             cfg=None,
             ),
    ]
)

stream4 = dict(
    type='radar_tracking_stream',
    description='雷达检测+跟踪流',
    models=[
        dict(type='radardet',
             cfg=None,
             ),
        dict(type='radarsort',
             cfg=None,
             ),
    ]
)

# 这里包含的才会初始化
"""
streams = [
    stream1,
    stream2,
    stream3,
    stream4,
    ]
"""
streams = []



"""
初始化注册，
在导入nn模块自动注册内部模块时，由于外部调用多次会导致对此初始化注册，仅初始化注册一次
"""
import damei as dm
import os
import sys
from pathlib import Path

logger = dm.get_logger('init_register')


class InitRegister(object):
    def __init__(self, internal_dir=None, api_fold_name='hai_api'):
        self.inited = False
        self.internal_dir = internal_dir  # /home/zzd/VSProject/hai/hai
        self.api_fold_name = api_fold_name
        self.show_logger = False

    def import_internal_module(self, internal_module_path):
        """
        导入内部模块
        :param internal_module_path: internal module path, e.g.: apis/modules/YOLOv5-v6
        """
        if self.internal_dir is None:
            full_path = os.path.join(os.getcwd(), internal_module_path)
        else:
            full_path = os.path.join(self.internal_dir, internal_module_path)  # /home/zzd/VSProject/hai/hai/apis/modules/YOLOv5_v6
        full_path = Path(full_path)
        # dir_path = str(full_path.parent)
        # stem = str(full_path.stem)

        imp = internal_module_path.split('/')
        imp = '.'.join(imp)

        # print('import_internal_module', imp)

        code = f'from {imp} import __init__'
        self.exec_import(code, internal_module_path)
        
    
    def import_external_module(self, external_module_path):
        """
        导入外部模块
        :param external_module_path: external module path, e.g.: repos/xxx/hai_api
        """
        emp = external_module_path
        if emp.startswith('~'):
            emp = emp.replace('~', str(Path.home()))
        emp = os.path.abspath(emp) 
        if not os.path.exists(os.path.join(emp, self.api_fold_name)):
            if self.show_logger:
                logger.warn(f'External module path "{emp}" not exists, and will be ignored.')
            return

        emp = os.path.join(emp, self.api_fold_name)
        # judge if the external module is a python package
        if not os.path.exists(os.path.join(emp, '__init__.py')):
            if self.show_logger:
                logger.warn(f'External module "{emp}" is not a python package, and will be ignored.')
            return
        # print('import_external_module', emp)
        
        self.import_init(emp)
                
        
    def import_init(self, path):
        # print(f'sys.path: {sys.path}')
        for p in sys.path:
            if p == '':
                continue
            elif p == path:
                package, name = '.', '__init__'
                code = f'from {package} import {name}'
                self.exec_import(code, package)
                return
            elif p in path:  # sys.path in input path, i.e. /home/zzd/VSProjects/hai in /home/zzd/VSProjects/hai/hai_api
                # print(f'sys.path in input path: p: {p} path: {path}')
                package = path.replace(p, '')
                package = '.'.join(package.split('/'))
                package = package[1:] if package.startswith('.') else package
                name = '__init__'
                code = f'from {package} import {name}'
                self.exec_import(code, package)
                return
            else:
                continue
        
        # print('import_init xx')
        # no env availabel i.e. /home/zzd/VSProjects/particle_transformer 
        # need add path to sys.path

        # Debug 01: Two modules with the same api_fold_name "hai_api", and only one will be registered.'
        #    Solve: change stem to project_name/stem, e.g. particle_transformer/hai_api, and the package name is shortten
        #    20220926: Unsolved.

        # dir = f'{Path(path).parent.parent}'
        dir = f'{Path(path).parent}'
        # dir = path
        stem = os.path.basename(path)
        cwd = os.getcwd()
        sys.path.insert(0, dir)
        os.chdir(dir)
        # package, name = f'{Path(path).parent.stem}/{stem}', '__init__'
        package, name = f'{stem}', '__init__'
        code = f'from {package} import {name}'
        # code = 'from . import *'
        # print('import_init', code, sys.path)
        self.exec_import(code, dir)
        os.chdir(cwd)  
        sys.path.pop(0)
        return

    def exec_import(self, code, path):
        """
        执行导入
        :param code: import code
        :param path: import path
        """
        if self.show_logger:
            logger.info(f'Internal module "{path}" registered.')
        exec(code)

    def __call__(self, internal_modules, external_folders):
        """
        初始化注册
        :param internal_modules: internal modules, e.g: models/loader/vis_loader
        :param external_folders: external folders, e.g.: dmapi, repos
        """
        if self.inited:
            pass
        else:
            # print(self.inited)
            ims = internal_modules
            efs = external_folders  # ['dmapi', 'repos']
            if len(ims) > 0:
                for i, im_path in enumerate(ims):
                    self.import_internal_module(im_path)

            if len(efs) > 0:
                # print(efs)
                # support external folders
                for i, ef in enumerate(efs):
                    self.import_external_module(ef)

                # dirs = [x for x in os.listdir('.') if os.path.isdir(x)]  # 当前路径文件夹
                # for dmp in dirs:
                #     if dmp in efs:
                #         # code = f'from repos.xxx.dmapi import *'
                #         code = f'from {dmp} import *'
                #         exec(code)
            self.inited = True






from typing import Any, List, Dict
from dataclasses import dataclass, field
from datetime import datetime
from ._related_class import WorkerInfo, UserInfo, APIKeyInfo


@dataclass
class HListPage:
    """用户Client从服务器端获取到List类的消息后解析该对象"""

    object: str = 'list'
    data: List[Any] = field(default_factory=list, metadata={'description': 'List of data'})
    first_id: str = None
    last_id: str = None
    has_more: bool = False

    def __len__(self):
        return len(self.data)
    
    def __iter__(self):
        return iter(self.data)
    
    def __getitem__(self, index):
        return self.data[index]
    

@dataclass
class HWorkerListPage(HListPage):
    """用户Client从服务器端获取到List类的消息后解析该对象"""
    data: List[WorkerInfo] = field(default_factory=List[WorkerInfo], metadata={'description': 'List of data'})

    def __post_init__(self):
        self.data = self.parse_worker_info(self.data)

    def parse_worker_info(self, data_list: List[dict]) -> List[WorkerInfo]:
        worker_infos = []
        for item in data_list:
            # Parse each item to populate WorkerInfo data class
            try:
                worker_info = WorkerInfo(**item)
                worker_infos.append(worker_info)
            except Exception as e:
                print(f"Failed to parse WorkerInfo: {e}")
                raise e
        return worker_infos
    
    def to_table(self):
        """转换为表格数据"""
        table_data = {
            "order": [],
            "id": [],
            "type": [],
            "worker_address": [],
            "model_name": [],
            "speed": [],
            "queue_length": [],
            "status": [],
            "model_onwer": [],
            "model_groups": [],
            "model_users": [],
            "host_name": [],
            "last_heartbeat": [],
            "model_functions": [],
            # "metadata": [],
        }
        for i, worker_info in enumerate(self.data):
            table_data["order"].append(i)
            table_data["id"].append(worker_info.id)
            table_data["type"].append(worker_info.type)
            table_data["worker_address"].append(worker_info.network_info.worker_address)
            table_data["model_name"].append(worker_info.resource_info.model_name)
            table_data["speed"].append(worker_info.status_info.speed)
            table_data["queue_length"].append(worker_info.status_info.queue_length)
            table_data["status"].append(worker_info.status_info.status)
            table_data["model_onwer"].append(worker_info.resource_info.model_onwer)
            table_data["model_groups"].append(worker_info.resource_info.model_groups)
            table_data["model_users"].append(worker_info.resource_info.model_users)
            table_data["model_functions"].append(worker_info.resource_info.model_functions)
            table_data["host_name"].append(worker_info.network_info.host_name)
            
            last_heartbeat = worker_info.last_heartbeat
            formatted_last_heartbeat = datetime.fromtimestamp(last_heartbeat).strftime("%Y-%m-%d %H:%M:%S")
            table_data["last_heartbeat"].append(formatted_last_heartbeat)
            # table_data["metadata"].append(worker_info.metadata)
        return table_data
    
@dataclass
class HUserListPage(HListPage):
    """用户Client从服务器端获取到List类的消息后解析该对象"""
    data: List[UserInfo] = field(default_factory=List[UserInfo], metadata={'description': 'List of data'})

    def __post_init__(self):
        # 解析用户数据
        self.data = self.parse_user_info(self.data)

    def parse_user_info(self, data_list: List[dict]) -> List[dict]:
        user_infos = []
        for item in data_list:
            user_info = UserInfo(**item)
            user_infos.append(user_info)
        return user_infos
    
    def to_table(self):
        table_data = {
            "order": [],
            "username": [],
            "user_level": [],
            "expiration_time": [],
            "auth_type": [],
            "user_groups": [],
            "phone": [],
            "email": [],
            "create_time": [],
            "id": [],
            "umt_id": [],
            "cluster_uid": [],
            "remarks": [],
        }
        for i, user_info in enumerate(self.data):
            table_data["order"].append(i)
            table_data["id"].append(user_info.id)
            table_data["username"].append(user_info.username)
            table_data["user_level"].append(user_info.user_level.str())
            table_data["expiration_time"].append(user_info.user_level.expiration_time)
            table_data["user_groups"].append([x.name for x in user_info.user_groups if x.name])
            table_data["phone"].append(user_info.phone)
            table_data["email"].append(user_info.email)
            table_data["auth_type"].append(user_info.auth_type)
            table_data["create_time"].append(user_info.create_time)
            table_data["remarks"].append(user_info.remarks)
            table_data["umt_id"].append(user_info.umt_id)
            table_data["cluster_uid"].append(user_info.cluster_uid)
        return table_data
    

@dataclass
class HAPIKeyListPage(HListPage):
    """用户Client从服务器端获取到List类的消息后解析该对象"""
    data: List[APIKeyInfo] = field(default_factory=List[APIKeyInfo], metadata={'description': 'List of data'})

    def __post_init__(self):
        self.data = self.parse_api_key_info(self.data)

    def parse_api_key_info(self, data_list: List[dict]) -> List[APIKeyInfo]:
        """将API-KEY信息解析为APIKeyInfo对象"""
        api_key_infos = []
        for item in data_list:
            if isinstance(item, dict):
                item = APIKeyInfo(**item)
            api_key_infos.append(item)
        return api_key_infos
    
    def to_table(self):
        table_data = {
            "order": [],
            "alias": [],
            "api_key": [],
            "expiration_time": [],
            "user_id": [],
            "id": [],
            "create_time": [],
            "create_by": [],
            "update_time": [],
            "remarks": [],
        }
        for i, api_key_info in enumerate(self.data):
            table_data["order"].append(i)
            table_data["id"].append(api_key_info.id)
            table_data["alias"].append(api_key_info.alias)
            table_data["api_key"].append(api_key_info.api_key)
            table_data["expiration_time"].append(api_key_info.expiration_time)
            table_data["create_time"].append(api_key_info.create_time)
            table_data["create_by"].append(api_key_info.created_by)
            table_data["update_time"].append(api_key_info.update_time)
            table_data["remarks"].append(api_key_info.remarks)
            table_data["user_id"].append(api_key_info.user_id)
        return table_data

    




import os, sys, signal
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union, Generator, Callable
import threading
import atexit
import time, asyncio
import traceback, inspect
from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.responses import StreamingResponse
import requests
import logging
from logging import Logger


from ._related_class import (
    WorkerNetworkInfo, ModelResourceInfo, WorkerStatusInfo, WorkerInfo,
    HWorkerModel, 
)

from .utils import get_uuid

@dataclass
class HWorkerArgs:  # (2) worker的参数配置和启动代码
    host: str = field(default="0.0.0.0", metadata={"help": "Worker's address, enable to access from outside if set to `0.0.0.0`, otherwise only localhost can access"})
    port: str = field(default="auto", metadata={"help": "Worker's port"})
    auto_start_port: int = field(default=42602, metadata={"help": "Worker's start port, only used when port is set to `auto`"})
    route_prefix: str = field(default="/apiv2", metadata={"help": "Route prefix for worker"})
    controller_address: str = field(default="http://localhost:42601", metadata={"help": "Controller's address"})
    controller_prefix: str = field(default="/apiv2", metadata={"help": "Controller's route prefix"})
    
    speed: int = field(default=1, metadata={"help": "Model's speed"})
    limit_model_concurrency: int = field(default=5, metadata={"help": "Limit the model's concurrency"})
    stream_interval: float = field(default=0., metadata={"help": "Extra interval for stream response"})
    no_register: bool = field(default=False, metadata={"help": "Do not register to controller"})
    permissions: str = field(default='groups: all; users: admin', metadata={"help": "Model's permissions, separated by ;, e.g., 'groups: all; users: a, b; owner: c'"})
    description: str = field(default='This is a demo worker of HaiDDF2 (HepAI Distrobuted Deployment Framework)', metadata={"help": "Model's description"})
    author: str = field(default=None, metadata={"help": "Model's author"})
    test: bool = field(default=False, metadata={"help": "Test mode, will not really start worker, just print the parameters"})



class CommonWorker:
    """
    Common Worker，用来实际处理各种函数
    """
    def __init__(self,
                 app: FastAPI,  # FastAPI app
                 model: HWorkerModel,  # 模型
                 worker_config: "HWorkerArgs",  # worker的配置
                 logger: Optional[Logger] = None,  # logger
                 **kwargs
                 ):
        # 配置
        self.config: HWorkerArgs = worker_config
        self.config_dict: Dict = self._init_config_dict(worker_config, **kwargs)
        
        # 网络信息相关项目
        self.network_info: WorkerNetworkInfo = self.get_network_info()
        
        # 模型资源信息相关
        self.model_permissions: Dict = self._init_permissions(self.config.permissions)  # 将str转换为dict

        # 状态信息相关
        self.app = app  # "HWorkerAPP" 
        self.speed = self.config_dict.get("speed", 1)
        self.status = "idle"

        # 初始化信息
        worker_id = self.config_dict.get("worker_id", None)
        self.worker_id = worker_id if worker_id else get_uuid(lenth=15, prefix="wk-")
        self.stream_interval = self.config_dict.get("stream_interval", 0)
        self.api_key = kwargs.get("api_key", "")
        self.model = model or HWorkerModel()

        self._is_deleted_in_controller = False  
        # 日志
        self.logger = logger or logging.getLogger(__name__)

        # 注册模型
        if not self.config.no_register:
            success: bool = self.register_to_controller()
            if success:
                self.heartbeat_thread = threading.Thread(
                    target=self.worker_heartbeat, 
                    daemon=True,
                    )
                self.heartbeat_thread.start()
            else:
                self._is_deleted_in_controller = True

        # 标识是否已经在controller中删除，如果已经删除，则exit_handler不再向controller发送删除信息
        # 用于适配controllre端主动向worker发送删除worker时
        atexit.register(self.exit_handler)

        self.status = "ready"

    @property
    def base_url(self):
        controller_addr = self.config.controller_address
        controller_prefix = self.config.controller_prefix
        if controller_prefix:
            return f'{controller_addr}{controller_prefix}'
        return controller_addr
    
    @property
    def worker_name(self):
        return self.model.name or self.model.__class__.__name__
    
    def _init_permissions(self, permissions: Optional[Union[str, Dict]]) -> Dict:
        """worker授予用户或者组的权限"""
        if permissions is None:
            return {}
        elif isinstance(permissions, dict):
            return permissions
        elif isinstance(permissions, str):
            """user: <user1>; user: <user2>; group: <group1>, ..."""
            prems = dict()
            for a in permissions.split(';'):
                user_or_group, name = a.split(':')
                user_or_group = user_or_group.strip()
                name = name.strip()
                assert user_or_group in ['owner', 'users', 'groups']
                names = name.split(',')  if ',' in name else [name]
                prems[user_or_group] = names
        else:
            raise ValueError(f"permissions should be str or dict, but got {type(permissions)}")
        return prems
    
    def _init_config_dict(self, config: HWorkerArgs, **kwargs):
        config_dict = config.__dict__
        for k, v in kwargs.items():
            if k in config_dict:
                raise ValueError(f"Duplicate key: {k}, Please check the config class and the kwargs.")
            config_dict[k] = v
        return config_dict
    
    def worker_heartbeat(self):
        """间隔发送心跳的函数"""
        heartbeat_interval = self.config_dict.get("heartbeat_interval", 60)
        while True:
            time.sleep(heartbeat_interval)
            self.register_to_controller(heartbeat_flag=True)
    
    def get_network_info(self) -> WorkerNetworkInfo:
        host = self.config.host
        from .utils import auto_port, auto_worker_address, get_hostname
        port = auto_port(port=self.config.port, start=self.config.auto_start_port)
        worker_address = auto_worker_address(worker_address="auto", host=host, port=port)
        route_prefix = self.config.route_prefix
        if route_prefix:
            worker_address = f"{worker_address}{route_prefix}"
        return WorkerNetworkInfo(
            host=self.config.host,
            port=port,
            host_name=get_hostname(),
            route_prefix=route_prefix,
            worker_address=worker_address,
        )
    
    def get_model_resource_info(self) -> ModelResourceInfo:
        model_name = self.model.name or self.model.__class__.__name__
        permmisions = self.model_permissions
        # callable_functions = self.get_callable_functions(self.model)
        callable_functions = self.get_all_callable_functions(self.model)
        return ModelResourceInfo(
            model_name=model_name,
            model_type=self.config_dict.get("model_type", "common"),
            model_version=self.config_dict.get("model_version", "1.0"),
            model_description=self.config.description,
            model_author=self.config.author,
            model_onwer=permmisions.get("owner", None),
            model_users=permmisions.get("users", []),
            model_groups=permmisions.get("groups", []),
            model_functions=callable_functions,
        )
    
    def get_callable_functions(self, cls):
        """获取一个类的所有可调用函数"""
        callable_funcs = []
        for name, method in inspect.getmembers(cls, inspect.isfunction):
            # Check if method belongs to the class and is not inherited
            if method.__qualname__.startswith(cls.__name__ + '.'):
                callable_funcs.append(name)
        return callable_funcs
    
    def get_all_callable_functions(self, cls: HWorkerModel):
        """获取一个类的所有可调用函数，包括父类的"""
        # Recursive function to collect callable methods from the class and its parent classes
        methods = {}
   
        # Traverse the method resolution order (MRO), which includes the class and its bases
        for subclass in inspect.getmro(cls.__class__):
            for name, method in inspect.getmembers(subclass, inspect.isfunction):
                methods[name] = method
        callable_funcs = list(methods.keys())

        # 再判断这些函数是不是可以远程调用的
        callable_funcs = [f for f in callable_funcs if hasattr(methods[f], "is_remote_callable")]

        return callable_funcs


    def get_status_info(self) -> WorkerStatusInfo:
        return WorkerStatusInfo(
            speed=self.speed,
            queue_length=self.app.get_queue_length(),
            status=self.status,
        )
    
    async def shutdown_worker(self) -> None:
        await asyncio.sleep(1)
        os.kill(os.getpid(), signal.SIGINT)
    
    def get_worker_info(self) -> WorkerInfo:
        worker_info = WorkerInfo(
            id=self.worker_id,
            network_info=self.network_info,
            resource_info=self.get_model_resource_info(),
            status_info=self.get_status_info(),
            check_heartbeat=True,
            vserion=self.config_dict.get("version", "2.0"),
            metadata={},
        )
        return worker_info
    
    def set_model_pressions(self, permissions: Dict):
        """设置模型的权限, 用于动态更新，仅限于owner和admin"""
        self.model_permissions = permissions

    def set_model_speed(self, speed: int):
        """设置模型的速度, 用于动态更新，仅限于owner和admin"""
        self.speed = speed
    
    ### --- 此处开始是向controller发送请求的相关函数 --- ###
    @property
    def headers(self) -> dict:
        headers = {
            'Authorization': f'Bearer {self.api_key}',
            'Content-Type': 'application/json'
        }
        return headers
    
    def register_to_controller(self, heartbeat_flag: bool = False):
        """注册和心跳"""
        # logger.info(f'Register model "{self.model_name}" to controller.')
        
        # url = self.base_url + '/register_worker'
        url = self.base_url + '/worker/register_worker'
        worker_info: WorkerInfo = self.get_worker_info()
        data = worker_info.to_dict()

        try:
            r = requests.post(url, json=data, headers=self.headers)
            if r.status_code != 200:
                raise Exception(f"Register worker failed. Error: {r.text}\n  Request url: {url}\n  Worker_info: {worker_info}")
            
        except Exception as e:
            if heartbeat_flag:
                print(f"Sent heartbeat failed, ignore... url: {url}, worker_id: {self.worker_id}")
                pass
            else:
                self.logger.warning(f"Register worker to controller failed, pass...")
                return False
        self.logger.info(f"Register worker successful")
        return True
  
    def exit_handler(self):
        """
        退出时向Contoller发送退出信息
        """
        if self.config.no_register:  # 初始化时没有注册到controller，不需要发送退出信息
            return
        if self._is_deleted_in_controller:
            return
        # logger.info(f'Remove model "{self.model_name}" from controller {self.controller_addr}, ')
        # url = self.base_url + "/stop_worker"
        url = self.base_url + "/worker/stop_worker"
        
        worker_info: WorkerInfo = self.get_worker_info()
        wid = worker_info.id
        payload = {"worker_id": wid, "permanent": False}
        # data = worker_info.to_dict()
        r = requests.post(
            url, 
            json=payload, 
            headers=self.headers)
        assert r.status_code == 200, f"Stop worker failed. {r.text}\n worker_info: {worker_info}"
        # logger.info(f'Done. {r.text}') 
        print(f"Stop worker successful. res: {r.text}")

    ### --- 此处是处理Controller的请求的相关函数 --- ###
    def unified_gate(
            self, 
            function: str,
            args,
            kwargs,
            ):
        """
        统一入口, v2.0更新为符合FastAPI规范，易于检测错误
        """
        # assert "function" in kwargs, "function is required"
        # function = kwargs.pop("function")
        
        has_function = hasattr(self.model, function)
        if not has_function:
            raise HTTPException(status_code=404, detail=f"Function `{function}` does not exist in the worker `{self.worker_name}`")
        is_callable = callable(getattr(self.model, function))
        is_remote_callable = hasattr(getattr(self.model, function), "is_remote_callable")
        if not is_callable:
            raise HTTPException(status_code=405, detail=f"Function `{function}` is not callable in the worker `{self.worker_name}`")
        if not is_remote_callable:
            raise HTTPException(status_code=405, detail=f"Function `{function}` is not remote callable in the worker `{self.worker_name}`, please add `@remote_callable` decorator to the function in worker model.")
        if has_function and is_callable and is_remote_callable:
            func: Callable = getattr(self.model, function)
            try:
                # res =  func(**kwargs)
                res = func(*args, **kwargs)
                if isinstance(res, Generator):
                    return StreamingResponse(res, media_type="application/octet-stream")
                return res
            except Exception as e:
                # 获取报错类型：e.__class__.__name__
                tb_str = traceback.format_exception(*sys.exc_info())
                tb_str = "".join(tb_str)
                # logger.debug(f"Error: {e}.\nTraceback: {tb_str}")
                e_class = e.__class__.__name__
                error_msg = e.__dict__.get("body", None)
                error_msg = error_msg if error_msg else f"{e_class}: {str(e)}"
                if e_class == "ModuleNotFoundError":
                    raise HTTPException(status_code=404, detail=error_msg)
                elif e_class == "NotImplementedError":
                    raise HTTPException(status_code=501, detail=error_msg)
                elif e_class == "BadRequestError":
                    raise HTTPException(status_code=400, detail=error_msg)
                elif e_class == "TypeError":
                    raise HTTPException(status_code=500, detail=error_msg)
                elif e_class == "APITimeoutError":
                    raise HTTPException(status_code=504, detail=error_msg)
                ## TODO: 其他报错类型转换为合适的报错状态码
                error_msg2 = f"{e_class}: {str(e)}"
                print(f"一种新的错误类型：{e_class}, 错误信息：{error_msg}\n{error_msg2}")
                raise HTTPException(status_code=400, detail=f'{error_msg2}\n{error_msg2}')
        else:
            raise HTTPException(status_code=404, detail=f"Function `{function}` does not exist or is not callable in the worker `{self.worker_name}`")
    




import os, sys
import cv2
import base64
import numpy as np  


def load_img(img_path):
    assert os.path.exists(img_path), f"图片{img_path}不存在"
    img = cv2.imread(img_path)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img

def sam_pre_process(data):
    
    input_points = data.pop('input_points', None)
    input_labels = data.pop('input_labels', None)
    input_boxes = data.pop('input_boxes', None)
    
    img = data.pop('img', None)
    assert img is not None, "The img need to be provided."
    if isinstance(img, str):  # 路径，加载为图片
        assert os.path.exists(img), f"图片{img}不存在"
        if os.path.isfile(img):
            img = load_img(img)
    if isinstance(img, np.ndarray):  # 图片，转换为base64
        img = cv2.imencode('.jpg', img)[1]
        img = str(base64.b64encode(img))[2:-1]
        
    
    if (input_points is None) and (input_boxes is None):
        auto_mask = True
    else:
        auto_mask = False
    
    if isinstance(input_points, np.ndarray):
        input_points = input_points.tolist()
    if input_points is not None and input_labels is None:
        input_labels = [1] * len(input_points)
    if isinstance(input_labels, np.ndarray):
        input_labels = input_labels.tolist()
    if isinstance(input_boxes, np.ndarray):
        input_boxes = input_boxes.tolist()
    
    messages = dict()
    messages['img'] = img
    messages['input_points'] = input_points
    messages['input_labels'] = input_labels
    messages['input_boxes'] = input_boxes
    messages['auto_mask'] = auto_mask
    
    new_data = dict()
    new_data['model'] = data['model']
    new_data['messages'] = messages
    new_data.update(data)
    
    return new_data

def sam_post_process(data):
    # return data
    if isinstance(data[0], dict):
        only_mask = False
    else:
        only_mask = True

    if only_mask:
        for mask_info in data:
            mask_info = np.array(mask_info, dtype=np.int32)
    else:
        for i, mask_info in enumerate(data):
            mask_info['segmentation'] = np.array(mask_info['segmentation'], dtype=np.int32)
        
    return data




import os, sys
from pathlib import Path
import hepai as hai
from fastapi import FastAPI
import uvicorn

try:
    from haiddf.version import __version__
except:
    sys.path.insert(1, str(Path(__file__).parent.parent.parent.parent))
    from haiddf.version import __version__

from haiddf._types import HWorkerModel
from haiddf.modules.worker.worker_app import HWorkerAPP, HWorkerArgs


def main():
    model = HWorkerModel()
    worker_config = hai.parse_args(HWorkerArgs)
    app: FastAPI = HWorkerAPP(model, worker_config=worker_config)

    print(f'worker_id: `{app.worker.worker_id}`', flush=True)
    
    uvicorn.run(app, host=app.host, port=app.port)

if __name__ == "__main__":
    main()



"""
这里是haiddf client类定义
"""
import os
import time
import httpx

from . import resources
from .resources._base_resource import SyncAPIResource
from .resources.worker_resource import HRemoteModel
from ._related_class import (
    HWorkerModel, WorkerInfo, WorkerStatusInfo, WorkerStoppedInfo,

)   


DEFAULT_API_KEY = os.getenv("DDF_API_KEY", None)

class HaiDDFClient(SyncAPIResource):
    key: resources.KeyManager
    http_client: httpx.Client

    def __init__(
        self,
        api_key: str | None = None,
        base_url: str | None = None,
        timeout: float | None = None,
        # max_retries: int = 0,
        proxy: str | None = None,
        print_result: bool = False,
        ):
        self.base_url = base_url if base_url is not None else f"http://localhost:42601/apiv2"
        self.api_key = api_key if api_key is not None else DEFAULT_API_KEY
        self.proxy = proxy
        self.timeout = timeout
        self.print_result = print_result
        super().__init__(client=self)

        self.key = resources.KeyManager(self)
        self.user = resources.UserResource(self)
        self.worker = resources.WorkerResource(self)

    ## --- 关于Worker的函数 --- ## 

    def list_workers(self):
        return self.worker.list()
    
    def request_worker(
            self,
            # target: dict,
            params: dict,
            args: list,
            kwargs: dict,
            ):
        """
        Request worker via http request

        Args:
            target (dict): Target worker and function, for example: {"model": "HWorkerModel", "function": "hello_world"}
            args (list): Arguments, can by any thing you defined in your `HWorkerModel`
            kwargs (dict): Keyword arguments, can by any thing you defined in your `HWorkerModel`
        """
        return self.worker.request(params=params, args=args, kwargs=kwargs)

    def register_worker(
            self,
            model: HWorkerModel,
            daemon: bool = False,
            standalone: bool = False,
            ):
        """
        注册Worker到服务端

        Args:
            model (HWorkerModel): Worker模型
            daemon (bool): 是否以守护进程方式启动
            standalone (bool): 是否以独立进程方式启动
        """
        return self.worker.register(model, daemon=daemon, standalone=standalone)

    def get_worker_info(
            self, 
            worker_id: str = None, 
            model_name: str = None,
            ) -> WorkerInfo:
        """
        Get worker info by worker_id via http request
    
        Args:
            worker_id (str): Worker ID
        """
        return self.worker.info(worker_id=worker_id, model_name=model_name)
    
    def get_worker_status(self, worker_id: str, refresh: bool = False) -> WorkerStatusInfo:
        """
        Get worker status by worker_id via http request
        
        Args:
            worker_id (str): Worker ID
            refresh (bool): Whether to refresh the worker status
        """
        return self.worker.status(worker_id=worker_id, refresh=refresh)

    def stop_worker(self, worker_id: str, permanent: bool = False) -> WorkerStoppedInfo:
        """
        Stop worker by worker_id via http request
        
        Args:
            worker_id (str): Worker ID
            permanent (bool): Whether to stop the worker permanently
        """
        return self.worker.stop(worker_id=worker_id, permanent=permanent)

    def refresh_all_workers(self) -> dict:
        """
        Refresh all workers
        """
        return self.worker.refresh_all()
    
    def get_remote_model(self, model_name: str) -> HRemoteModel:
        """
        Get a remote model
        """
        return self.worker.get_remote_model(model_name=model_name)
    
    ## # --- 关于User的函数 --- ##
    def list_users(self):
        """
        List all users
        Note: Only admin can use this function
        """
        return self.user.list_users()
    
    def create_user(self, **kwargs):
        """
        Create a new user
        Note: Only admin can use this function
        """
        return self.user.create_user(**kwargs)
    
    def delete_user(self, user_id: str):
        """
        Delete a user
        Note: Only admin can use this function
        """
        return self.user.delete_user(user_id=user_id)
    
    ### --- 关于Key的函数 --- ###
    def list_api_keys(self):
        """
        List all keys
        Note: Only admin can use this function
        """
        return self.key.list_api_keys()
    
    def create_api_key(self, **kwargs):
        """
        Create a new key
        Note: Only admin can use this function
        """
        return self.key.create_api_key(**kwargs)
    
    def delete_api_key(self, api_key_id: str):
        """
        Delete a key
        Note: Only admin can use this function
        """
        return self.key.delete_api_key(api_key_id=api_key_id)

if __name__ == "__main__":
    pass

    







import argparse
import logging
import os

import numpy as np
import torch
import torch.nn.functional as F
from PIL import Image
from torchvision import transforms

from utils.data_loading import BasicDataset
from unet import UNet
from utils.utils import plot_img_and_mask

def predict_img(net,
                full_img,
                device,
                scale_factor=1,
                out_threshold=0.5):
    net.eval()
    img = torch.from_numpy(BasicDataset.preprocess(full_img, scale_factor, is_mask=False))
    img = img.unsqueeze(0)
    img = img.to(device=device, dtype=torch.float32)

    with torch.no_grad():
        output = net(img)

        if net.n_classes > 1:
            probs = F.softmax(output, dim=1)[0]
        else:
            probs = torch.sigmoid(output)[0]

        tf = transforms.Compose([
            transforms.ToPILImage(),
            transforms.Resize((full_img.size[1], full_img.size[0])),
            transforms.ToTensor()
        ])

        full_mask = tf(probs.cpu()).squeeze()

    if net.n_classes == 1:
        return (full_mask > out_threshold).numpy()
    else:
        return F.one_hot(full_mask.argmax(dim=0), net.n_classes).permute(2, 0, 1).numpy()


def get_args():
    parser = argparse.ArgumentParser(description='Predict masks from input images')
    parser.add_argument('--model', '-m', default='MODEL.pth', metavar='FILE',
                        help='Specify the file in which the model is stored')
    parser.add_argument('--input', '-i', metavar='INPUT', nargs='+', help='Filenames of input images', required=True)
    parser.add_argument('--output', '-o', metavar='OUTPUT', nargs='+', help='Filenames of output images')
    parser.add_argument('--viz', '-v', action='store_true',
                        help='Visualize the images as they are processed')
    parser.add_argument('--no-save', '-n', action='store_true', help='Do not save the output masks')
    parser.add_argument('--mask-threshold', '-t', type=float, default=0.5,
                        help='Minimum probability value to consider a mask pixel white')
    parser.add_argument('--scale', '-s', type=float, default=0.5,
                        help='Scale factor for the input images')
    parser.add_argument('--bilinear', action='store_true', default=False, help='Use bilinear upsampling')

    return parser.parse_args()


def get_output_filenames(args):
    def _generate_name(fn):
        return f'{os.path.splitext(fn)[0]}_OUT.png'

    return args.output or list(map(_generate_name, args.input))


def mask_to_image(mask: np.ndarray):
    if mask.ndim == 2:
        return Image.fromarray((mask * 255).astype(np.uint8))
    elif mask.ndim == 3:
        return Image.fromarray((np.argmax(mask, axis=0) * 255 / mask.shape[0]).astype(np.uint8))


if __name__ == '__main__':
    args = get_args()
    in_files = args.input
    out_files = get_output_filenames(args)

    net = UNet(n_channels=3, n_classes=2, bilinear=args.bilinear)

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    logging.info(f'Loading model {args.model}')
    logging.info(f'Using device {device}')

    net.to(device=device)
    net.load_state_dict(torch.load(args.model, map_location=device))

    logging.info('Model loaded!')

    for i, filename in enumerate(in_files):
        logging.info(f'\nPredicting image {filename} ...')
        img = Image.open(filename)

        mask = predict_img(net=net,
                           full_img=img,
                           scale_factor=args.scale,
                           out_threshold=args.mask_threshold,
                           device=device)

        if not args.no_save:
            out_filename = out_files[i]
            result = mask_to_image(mask)
            result.save(out_filename)
            logging.info(f'Mask saved to {out_filename}')

        if args.viz:
            logging.info(f'Visualizing results for image {filename}, close to continue...')
            plot_img_and_mask(img, mask)



"""
自定义模型，由HaiDDF搭载，实现远程服务和调用
"""
from dataclasses import dataclass, field
import uvicorn

import hepai as hai
# from hepai._types import HWorkerModel, HWorkerArgs, HWorkerAPP

try:
    from haiddf.version import __version__
except:
    from pathlib import Path
    here = Path(__file__).parent
    import os, sys
    sys.path.insert(1, str(here.parent.parent.parent))
    from haiddf.version import __version__
from haiddf._types import HWorkerModel, HWorkerArgs
from worker_app import HWorkerAPP

# 继承HWorkerModel类，实现自定义类
class CustomWorkerModel(HWorkerModel):
    def __init__(self, name: str = "hepai/custom-model"):
        super().__init__(name=name)

    @HWorkerModel.remote_callable  # 用该修饰器标记为可远程调用的方法
    def custom_method(self, a: int, b: int) -> int:
        """你可以在这里定义你的自定义方法和返回值"""
        return a + b

@dataclass
class CustomWorkerArgs(HWorkerArgs):
    """
    自定义参数类，继承自HWorkerArgs，
        - 可在此处定义自己的默认参数，添加新的参数
        - 也可以在程序运行时通过命令行传入参数，例如：--no_register True
    """
    pass
    
if __name__ == "__main__":
    # 实例化模型
    model = CustomWorkerModel(name="hepai/custom-model")
    # 解析命令行参数
    worker_config = hai.parse_args(CustomWorkerArgs)
    # 实例化APP，是一个fastapi应用
    app = HWorkerAPP(model, worker_config=worker_config)
    print(app.worker.get_worker_info(), flush=True)
    # 启动服务
    uvicorn.run(app, host=app.host, port=app.port)






import torch
from torch import Tensor


def dice_coeff(input: Tensor, target: Tensor, reduce_batch_first: bool = False, epsilon=1e-6):
    # Average of Dice coefficient for all batches, or for a single mask
    assert input.size() == target.size()
    if input.dim() == 2 and reduce_batch_first:
        raise ValueError(f'Dice: asked to reduce batch but got tensor without batch dimension (shape {input.shape})')

    if input.dim() == 2 or reduce_batch_first:
        inter = torch.dot(input.reshape(-1), target.reshape(-1))
        sets_sum = torch.sum(input) + torch.sum(target)
        if sets_sum.item() == 0:
            sets_sum = 2 * inter

        return (2 * inter + epsilon) / (sets_sum + epsilon)
    else:
        # compute and average metric for each batch element
        dice = 0
        for i in range(input.shape[0]):
            dice += dice_coeff(input[i, ...], target[i, ...])
        return dice / input.shape[0]


def multiclass_dice_coeff(input: Tensor, target: Tensor, reduce_batch_first: bool = False, epsilon=1e-6):
    # Average of Dice coefficient for all classes
    assert input.size() == target.size()
    dice = 0
    for channel in range(input.shape[1]):
        dice += dice_coeff(input[:, channel, ...], target[:, channel, ...], reduce_batch_first, epsilon)

    return dice / input.shape[1]


def dice_loss(input: Tensor, target: Tensor, multiclass: bool = False):
    # Dice loss (objective to minimize) between 0 and 1
    assert input.size() == target.size()
    fn = multiclass_dice_coeff if multiclass else dice_coeff
    return 1 - fn(input, target, reduce_batch_first=True)



import os
import time
import socketio

debug = False
sio = socketio.Client(logger=debug, engineio_logger=debug)


@sio.event
def connect():
    print('connection established')


@sio.event
def on_process(*args):
    print(f'on process: {len(args)} args: {args}')
    return args


@sio.event
def my_message(data):
    print('client 收到my_massage函数:', data)
    sio.emit('server_run_func', {'response': data}, callback=on_process)
    return 'xx'


@sio.event
def disconnect():
    print('disconnected from server')


# sio.connect('http://localhost:5000')
sio.connect('http://127.0.0.1:5000')

for i in range(10):
    data = {'a': f'{i}'}
    ret = my_message(data)
    print(f'ret: {ret}')
    time.sleep(0.5)

sio.wait()





from typing import Literal, Optional, List, Dict
from ._base_resource import SyncAPIResource
from .._client_class import HUserListPage
from .._related_class import UserInfo, UserDeletedInfo

class UserResource(SyncAPIResource):

    @property
    def path(self):
        return '/user'

    def list_users(self):
        url = f'{self.base_url}{self.path}/list_users'
        return self._get(url, headers=self.headers, cast_to=HUserListPage)
    
    def create_user(self, **kwargs):
        url = f'{self.base_url}{self.path}/create_user'
        return self._post(
            url, 
            headers=self.headers, 
            json=kwargs,
            cast_to=UserInfo,
            )
    
    def delete_user(self, user_id: str):
        url = f'{self.base_url}{self.path}/delete_user'
        payload = {
            "user_id": str(user_id),
        }
        return self._post(
            url, 
            headers=self.headers, 
            json=payload,
            cast_to=UserDeletedInfo,
            )

    ## --- 以下的LiteLLM的，准备弃用 --- ##
    def list_customers(self):
        """
        curl --location --request GET 'http://0.0.0.0:4000/customer/list'         --header 'Authorization: Bearer sk-1234'
        """
        url = f'{self.base_url}/customer/list'
        headers = self.headers
        return self._get(url, headers=headers)
    
    def create_customer(
            self,
            user_id: str,
            alias: Optional[str] = None,
            blocked: bool = False,
            max_budget: Optional[float] = None,
            budget_id: Optional[str] = None,
            allowed_model_region: Optional[Literal["eu"]] = None,
            default_model: Optional[str] = None,
            metadata: Optional[dict] = None,
            ):
        """
        user_id: str - The unique identifier for the user.
        alias: Optional[str] - A human-friendly alias for the user.
        blocked: bool - Flag to allow or disallow requests for this end-user. Default is False.
        max_budget: Optional[float] - The maximum budget allocated to the user. Either 'max_budget' or 'budget_id' should be provided, not both.
        budget_id: Optional[str] - The identifier for an existing budget allocated to the user. Either 'max_budget' or 'budget_id' should be provided, not both.
        allowed_model_region: Optional[Literal["eu"]] - Require all user requests to use models in this specific region.
        default_model: Optional[str] - If no equivalent model in the allowed region, default all requests to this model.
        metadata: Optional[dict] = Metadata for customer, store information for customer. Example metadata = {"data_training_opt_out": True}
        
        curl:
            curl --location 'http://0.0.0.0:4000/customer/new'         --header 'Authorization: Bearer sk-1234'         --header 'Content-Type: application/json'         --data '{
            "user_id" : "ishaan-jaff-3",
            "allowed_region": "eu",
            "budget_id": "free_tier",
            "default_model": "azure/gpt-3.5-turbo-eu" <- all calls from this user, use this model? 
        }'

        # return end-user object
        """
        url = f'{self.base_url}/customer/new'
        headers = self.headers
        data = {
            "user_id": user_id,
            "alias": alias,
            "blocked": blocked,
            "max_budget": max_budget,
            "budget_id": budget_id,
            "allowed_model_region": allowed_model_region,
            "default_model": default_model,
            "metadata": metadata,
        }
        return self._post(url, headers=headers, json=data)

    def create(
            self,
            user_id: Optional[str] = None,
            user_alias: Optional[str] = None,
            teams: Optional[List[str]] = [],
            organization_id: Optional[str] = None,
            user_email: Optional[str] = None,
            send_invite_email: Optional[bool] = False,
            user_role: Literal["proxy_admin", "proxy_admin_viewer", "internal_user", "internal_user_viewer", "team", "customer"] = "customer",
            max_budget: Optional[float] = None,
            budget_duration: Optional[Literal["30s", "30m", "30h", "30d"]] = None,
            models: Optional[List[str]] = [],
            tpm_limit: Optional[int] = None,
            rpm_limit: Optional[int] = None,
            auto_create_key: bool = True,
            ):
        """
        Use this to create a new INTERNAL user with a budget. Internal Users can access LiteLLM Admin UI to make keys, request access to models. This creates a new user and generates a new api key for the new user. The new api key is returned.
        Args:
            user_id: Optional[str] - Specify a user id. If not set, a unique id will be generated.
            user_alias: Optional[str] - A descriptive name for you to know who this user id refers to.
            teams: Optional[list] - specify a list of team id's a user belongs to.
            organization_id: Optional[str] - specify the org a user belongs to.
            user_email: Optional[str] - Specify a user email.
            send_invite_email: Optional[bool] - Specify if an invite email should be sent.
            user_role: Optional[str] - Specify a user role - "proxy_admin", "proxy_admin_viewer", "internal_user", "internal_user_viewer", "team", "customer". Info about each role here: https://github.com/BerriAI/litellm/litellm/proxy/_types.py#L20
            max_budget: Optional[float] - Specify max budget for a given user.
            budget_duration: Optional[str] - Budget is reset at the end of specified duration. If not set, budget is never reset. You can set duration as seconds ("30s"), minutes ("30m"), hours ("30h"), days ("30d").
            models: Optional[list] - Model_name's a user is allowed to call. (if empty, key is allowed to call all models)
            tpm_limit: Optional[int] - Specify tpm limit for a given user (Tokens per minute)
            rpm_limit: Optional[int] - Specify rpm limit for a given user (Requests per minute)
            auto_create_key: bool - Default=True. Flag used for returning a key as part of the /user/new response
        Returns:
            key: (str) The generated api key for the user
            expires: (datetime) Datetime object for when key expires.
            user_id: (str) Unique user id - used for tracking spend across multiple keys for same user id.
            max_budget: (float|None) Max budget for given user.
        """

        url = f'{self.base_url}/user/new'
        headers = self.headers
        data = {
            "user_id": user_id,
            "user_alias": user_alias,
            "teams": teams,
            "organization_id": organization_id,
            "user_email": user_email,
            "send_invite_email": send_invite_email,
            "user_role": user_role,
            "max_budget": max_budget,
            "budget_duration": budget_duration,
            "models": models,
            "tpm_limit": tpm_limit,
            "rpm_limit": rpm_limit,
            "auto_create_key": auto_create_key,
        }
        return self._post(url, headers=headers, json=data)



#!coding:utf-8

"""
Segment via HepAI based on SAM.
Author: JiaMeng Zhao, Zhengde Zhang
All rights reserved, 2023. 
This script is free for non-commercial use. Please contact zdzhang@ihep.ac.cn for commercial use.
"""

# from utils import *
import cv2
import os, sys
from pathlib import Path
here = Path(__file__).parent
import damei as dm
import numpy as np
import base64
import requests
logger = dm.get_logger('seg_via_sam.py')


class SegmentViaSam():

    def __init__(self):
        # 加载模型
        self.url = "http://aiapi.ihep.ac.cn:42901/v1/inference"
        self.model_name = "meta/segment_anything_model"
        self.api_key = os.environ.get("HEPAI_API_KEY")
        # assert api_key is not None, "请设置环境变量HEPAI_API_KEY"
        
    def load_img(self, img_path):
        assert os.path.exists(img_path), f"图片{img_path}不存在"
        img = cv2.imread(img_path)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return img

    def inference(self, img, **kwargs):
        logger.info(f"Inference ...")
        url = kwargs.get('url', self.url)
        model_name = kwargs.get('model_name', self.model_name)
        stream = kwargs.get('stream', False)
       
        input_points = kwargs.get('input_points', None)
        input_labels = kwargs.get('input_labels', None)
        input_boxes = kwargs.get('input_boxes', None)
         
        # auto_mask = kwargs.get('auto_mask', True)
        if (input_points is None) and (input_boxes is None):
            auto_mask = True
        else:
            auto_mask = False
        
        if isinstance(img, str):  # 路径，加载为图片
            if os.path.isfile(img):
                img = self.load_img(img)
        if isinstance(img, np.ndarray):  # 图片，转换为base64
            img = cv2.imencode('.jpg', img)[1]
            img = str(base64.b64encode(img))[2:-1]
            
        if isinstance(input_points, np.ndarray):
            input_points = input_points.tolist()
        if input_points is not None and input_labels is None:
            input_labels = [1] * len(input_points)
        if isinstance(input_labels, np.ndarray):
            input_labels = input_labels.tolist()
        if isinstance(input_boxes, np.ndarray):
            input_boxes = input_boxes.tolist()
            
        messages = dict()
        messages['img'] = img
        messages['input_points'] = input_points
        messages['input_labels'] = input_labels
        messages['input_boxes'] = input_boxes
        messages['auto_mask'] = auto_mask
        
        response = requests.post(
            url,
            headers={"Authorization": f"Bearer {self.api_key}"},
            json={
                "model": model_name,
                "api_key": self.api_key,
                "stream": stream,
                "messages": messages,
            },
            timeout=60,
            stream=stream
        )
        if response.status_code != 200:
            raise Exception(f"Got status code {response.status_code} from server: {response.text}")
            # print(response.content)  # 只有非流式相应才能查看内容
        data = response.json()
        if data['status_code'] != 42901:
            error_info = f'Request error: {data}'
            logger.error(error_info)
            raise NameError(error_info)
            return error_info
        masks = data['message']
        if auto_mask:  # 全景分割
            return self.post_prosscess_of_panorama(masks, **kwargs)
        return masks

    def post_prosscess_of_panorama(self, masks, **kwargs):
        return_rich_masks = kwargs.get('return_rich_masks', True)
        if return_rich_masks:
            return masks


def seg_imgs(svs):
    img_dir = f'{here.parent.parent}/imgs/008_ImM1_cracks'
    
    imgs = sorted(os.listdir(img_dir))
    for img in imgs:
        img = f'{img_dir}/{img}'
        masks = svs.inference(img)
        print(masks)
        
    
        exit()
    

if __name__ == '__main__':
    svs = SegmentViaSam()
    
    seg_imgs(svs)
    exit()
   
    
    fold = f'xiaoshunao2_output'
    img = f'{here}/imgs/{fold}/raw.png'
    img = f'{here}/imgs/{fold}/thresh_img&img.png'  # 146
    # img = f'{here}/imgs/{fold}/equ_img&img.png'   # 17
    # img = f'{here}/imgs/{fold}/thresh&equ&img.png'  # 8
    img = svs.load_img(img_path=img)
    masks = svs.inference(
        img, 
        # input_points=[[100, 100]],
        )
    # for mask in masks:
    #     segmentation = mask['segmentation']
    #     mask = np.array(segmentation)
    #     print(masks)
    # maskss = np.array(masks)  # shape: (1, 922, 964, 4)
    # masks = maskss[0]
    # tmp = None
    # for i in range(maskss.shape[-1]):  # 4ci 
    #     mask = masks[:, :, i]
    #     cv2.imshow(f'mask{i}', mask)
    #     cv2.waitKey(0)
    #     if tmp is not None:
    #         differ = np.sum(mask-tmp)
    #         print(f'differ: {i} {differ}')
    #     tmp = mask
    # print(masks)
    

    plt.figure(figsize=(10, 10))
    plt.imshow(img)
    show_anns(masks, plt.gca())
    # show_bboxs(mask, plt.gca())
    plt.axis('off')
    plt.show()


import os
import time
import httpx
from typing import Union, Optional, Dict
from functools import wraps
from dataclasses import dataclass, field

@dataclass
class HErrorResponse(Exception):
    """Class to encapsulate HTTP request errors."""
    status_code: int
    message: str
    content: Optional[str] = None

    def __str__(self):
        details = f"HTTP Status: {self.status_code} - {self.message}"
        if self.content:
            details += f"\nContent: {self.content}"
        return details

from ..utils import post_process_decorator

@dataclass
class HClient:
    """基础的客户端类"""
    api_key: str = field(default=None, metadata={'description': 'API Key'})
    base_url: str = field(default=None, metadata={'description': 'Base URL'})
    proxy: str = field(default=None, metadata={'description': 'Proxy URL'})
    timeout: float = field(default=600.0, metadata={'description': 'Timeout in seconds'})

    def __post_init__(self):
        self.sync_api_resource = SyncAPIResource(client=self)
        
        self._post = self.sync_api_resource._post
        self._get = self.sync_api_resource._get
        self._put = self.sync_api_resource._put
        self._delete = self.sync_api_resource._delete
        self._patch = self.sync_api_resource._patch
        self._head = self.sync_api_resource._head
        self._options = self.sync_api_resource._options


class SyncAPIResource:

    def __init__(
            self, 
            client = None,   # HaiDDFClient
            ):
        if isinstance(client, dict):
            client = HClient(**client)
        if isinstance(client, HClient):
            self._client = client
        else:
            # HaiDDFClient
            from ..haiddf_client import HaiDDFClient
            self._client: HaiDDFClient = client
        self.api_key = self._client.api_key
        self.base_url = self._client.base_url
        self.proxy = self._client.proxy
        self.timeout = self._client.timeout
        self._http_client: httpx.Client = None


    @property
    def http_client(self) -> httpx.Client:
        if self._http_client is None:
            _client = self.get_http_client(
                proxy=self.proxy,
                timeout=self.timeout,
                base_url=self.base_url,
                )
            self._http_client = _client
        return self._http_client

    @post_process_decorator
    def _get(
        self, 
        url: str,
        *,
        headers = None,
        params = None,
        **kwargs):
        if params is not None:
            kwargs.update({
                # "headers": headers, 
                "params": params})
        return self.http_client.get(url, headers=headers, **kwargs)
        # return self.http_client.get(*args, **kwargs)
    
    @post_process_decorator
    def _post(
            self, 
            url: str,
            *,
            headers = None,
            json = None,
            params = None,
            **kwargs):
        headers = headers if headers else self.headers
        # kwargs.update({"headers": headers, "json": json})
        return self.http_client.post(
            url, 
            json=json,
            headers=headers,
            params=params,
            **kwargs)
        # return self.http_client.post(*args, **kwargs)
    
    def _put(self, *args, **kwargs):
        return self.http_client.put(*args, **kwargs)
    
    def _delete(self, *args, **kwargs):
        return self.http_client.delete(*args, **kwargs)
    
    def _patch(self, *args, **kwargs):
        return self.http_client.patch(*args, **kwargs)
    
    def _head(self, *args, **kwargs):
        return self.http_client.head(*args, **kwargs)
    
    def _options(self, *args, **kwargs):
        return self.http_client.options(*args, **kwargs)

    @classmethod
    def get_http_client(cls, **kwargs) -> httpx.Client:
        DEFAULT_TIMEOUT = httpx.Timeout(timeout=600.0, connect=5.0)
        DEFAULT_LIMITS = httpx.Limits(max_connections=100, max_keepalive_connections=20)
        proxy = kwargs.get("proxy", None)
        if proxy is None:
            proxies = None
        else:
            proxies = {
                "http://": proxy,
                "https://": proxy,
            }
        base_url = kwargs.get("base_url", None)
        base_url = base_url or os.getenv("DDF_BASE_URL")
        timeout = kwargs.get("timeout", DEFAULT_TIMEOUT)
        timeout = DEFAULT_TIMEOUT if (timeout is None) else timeout
        transport = kwargs.get("transport", None)
        limits = kwargs.get("limits", DEFAULT_LIMITS)
        limits = DEFAULT_LIMITS if (limits is None) else limits
        http_client = httpx.Client(
            base_url=base_url,
            timeout=timeout,
            proxies=proxies,
            transport=transport,
            limits=limits,
        )
        return http_client

    
    @property
    def headers(self) -> dict:
        # if self.api_key is None:
        #     raise ValueError("API Key is required.")
        headers = dict()
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        headers["Content-Type"] = "application/json"
        return headers
    
    def _sleep(self, seconds: float) -> None:
        time.sleep(seconds)



"""
抽象的队列
"""

import os, sys

from collections import deque


class AbstractQue(deque):

    def __init__(self, maxlen=5, *args, **kwargs):
        super(AbstractQue, self).__init__(maxlen=maxlen, *args, **kwargs)

    def get_last(self):
        item = self.popleft()
        while self:
            item = self.popleft()
        return item







from typing import Dict, List, Literal, Optional, Union, Generator, Callable
from dataclasses import dataclass, field

from fastapi import FastAPI, APIRouter, Request, HTTPException
from fastapi.background import BackgroundTasks
from fastapi.responses import JSONResponse, StreamingResponse, FileResponse, HTMLResponse
from fastapi.params import Query

import asyncio
from asyncio import Semaphore
import threading
from pydantic import BaseModel
from logging import Logger

from ._worker_class import HWorkerArgs, HWorkerModel, CommonWorker
from ._related_class import WorkerStoppedInfo, WorkerInfoRequest, WorkerUnifiedGateRequest


class FunctionParamsItem(BaseModel):
    args: List = []
    kwargs: Dict = {}


class HWorkerAPP(FastAPI):
    """
    FastAPI app for worker
    """

    def __init__(
            self,
            model: HWorkerModel,
            worker_config: HWorkerArgs = None,
            logger: Logger = None,
            **kwargs):
        super().__init__()
        worker_config = worker_config if worker_config is not None else HWorkerArgs()
        # 用于控制模型访问的信号量
        self.limit_model_concurrency = worker_config.limit_model_concurrency
        self.model_semaphore: Semaphore = None
        self.global_counter = 0

        self.logger = self.get_logger(logger)
        self.worker = CommonWorker(
            app=self, model=model, worker_config=worker_config, 
            logger=self.logger,
            **kwargs)
        self._init_routers()

    def _init_routers(self):
        
        router_prefix = self.worker.config_dict.get("route_prefix", "")
        index_router = APIRouter(prefix="", tags=["base"])
        index_router.get("/")(self.index)
        self.include_router(index_router)

        router = APIRouter(prefix=router_prefix, tags=["worker"])
        router.post("/worker_unified_gate/")(self.worker_unified_gate)
        router.post("/worker_unified_gate/{function}")(self.worker_unified_gate)
        router.get("/worker_get_status")(self.worker_get_status)
        router.post("/shutdown_worker")(self.shutdown_worker)
        router.post("/worker/get_worker_info")(self.get_worker_info)  # 这个路由是为了与controller相同的格式，使得client也能调用
        router.post("/worker/unified_gate")(self.worker_unified_gate)  # 这个路由是为了与controller相同的格式，使得client也能调用

        self.include_router(router)

    @classmethod
    def get_logger(cls, logger: Logger = None):
        if logger is None:
            try:
                from ...utils._logger import Logger
                logger = Logger.get_logger("worker_app.py")
            except:
                import logging
                logger = logging.getLogger("worker_app.py")
        return logger

    @property
    def host(self):
        return self.worker.network_info.host
    
    @property
    def port(self):
        return self.worker.network_info.port

    def get_queue_length(self):
        model_semaphore: asyncio.Semaphore = self.model_semaphore
        if model_semaphore is None:
            return 0
        else:
            _value = model_semaphore._value
            _num_waiters = len(model_semaphore._waiters) if model_semaphore._waiters is not None else 0
            return self.limit_model_concurrency - _value + _num_waiters
    
    def release_model_semaphore(self):
        if self.model_semaphore is not None:
            self.model_semaphore.release()

    async def index(self):
        worker_name = self.worker.worker_name
        content = f"""
        <html>
            <head>
                <title>HepAI Worker Info</title>
            </head>
            <body>
                <h1>This is a worker of HepAI Distributed Deployment Framework</h1>
                <p>Worker Name: <strong>{worker_name}</strong></p>
                <p>Visit the <a href="/docs">API Documentation</a>.</p>
            </body>
        </html>
        """
        return HTMLResponse(content=content)
    
    # async def local_worker_unified_gate(
    #         self,
    #         worker_unified_gate_request: WorkerUnifiedGateRequest,
    #         model: str,
    #         function: str = "__call__",
    #         # user_auth: HAPIKeyAuth = api_key_auth,
    #         ) -> JSONResponse:
    #     wk_ugr = worker_unified_gate_request
    #     return self.worker.worker_unified_gate()


    async def worker_unified_gate(
            self,
            function_params: FunctionParamsItem,
            function: str = "__call__",
            ):
        # global model_semaphore, global_counter
        model_semaphore = self.model_semaphore  # 这个是用于获取队列长度的
        global_counter = self.global_counter
        global_counter += 1
        if model_semaphore is None:
            model_semaphore = asyncio.Semaphore(
                self.limit_model_concurrency)
        await model_semaphore.acquire()
        try:
            rst = self.worker.unified_gate(
                function=function, 
                args=function_params.args,
                kwargs=function_params.kwargs,)
        except Exception as e:  # 如果出错，也要释放锁
            self.release_model_semaphore()
            raise e
        # 如果成功，添加背景任务并返回
        background_tasks = BackgroundTasks()  # 背景任务
        background_tasks.add_task(self.release_model_semaphore)  # 释放锁
        return rst
    
    async def worker_get_status(self):
        return self.worker.get_status_info().to_dict()
    
    async def get_worker_info(
            self, 
            worker_info_request: WorkerInfoRequest,
            # user_auth: HAPIKeyAuth = api_key_auth,
            ) -> JSONResponse:
        """
        与controller的worker_info接口一致，以便client调用
        """
        return self.worker.get_worker_info()
    
    
    async def shutdown_worker(self, background_tasks: BackgroundTasks):
        """接收来自Ctrl的关闭worker的信息"""
        background_tasks.add_task(self.worker.shutdown_worker)
        self.worker._is_deleted_in_controller = True
        wid = self.worker.worker_id
        return WorkerStoppedInfo(
            id=wid, stopped=True, 
            message=f"Worker `{wid}` shutdown",
            shutdown=True,
            )

    @classmethod
    def register_worker(
            cls,
            model: HWorkerModel,
            worker_config: HWorkerArgs = None,
            daemon: bool = False,
            standalone: bool = False,
            **kwargs,
            ):
        """注册HModelWorker到HaiDDF"""
        from .utils import run_standlone_worker_demo
        if standalone:  # 独立程序模式，用户测试程序
            # cls().logger.info("Running `worker` in standalone mode.")
            print("Running `worker` in standalone mode.")
            return run_standlone_worker_demo()
        
        import uvicorn
        app: FastAPI = HWorkerAPP(model, worker_config=worker_config, **kwargs)
        def run_uvicron():
            uvicorn.run(app, host=app.host, port=app.port)

        if daemon:  # 守护进程模式，app workre在后台运行
            t = threading.Thread(target=run_uvicron, daemon=True)
            t.start()
            return app.worker.get_worker_info()
        else:  # 正常模式，app worker在前台运行
            run_uvicron()










"""
把所有client模块以外从外部引用的类定义引用到这里，方便把haiddf拷贝到hepai项目中使用
"""


# from ...controller.worker_manager._worker_class import (
#     HWorkerModel, WorkerInfo, WorkerStatusInfo, WorkerStoppedInfo,
#     )

# from ...controller.user_manager._user_class import (
#     UserInfo, UserDeletedInfo, APIKeyInfo, APIKEYDeletedInfo,
#     )


### --- 为HepAI --- ###
from ..base_class._worker_class import (
    HWorkerModel, WorkerInfo, WorkerStatusInfo, WorkerStoppedInfo,
    )

from ..base_class._user_class import (
    UserInfo, UserDeletedInfo, APIKeyInfo, APIKEYDeletedInfo,
    )




def get_uuid(lenth, prefix=None):
    import uuid
    if prefix:
        return prefix + str(uuid.uuid4())[:lenth-len(prefix)]
    return str(uuid.uuid4())[:lenth]

def get_hostname():
    import socket
    return socket.gethostname()


def get_used_ports(start=1, end=65535):
    import socket
    ports = []
    for port in range(start, end):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind(('0.0.0.0', port))
        except:
            ports.append(port)
        finally:
            s.close()
    return ports


def auto_port(port=None, start=42901, **kwargs):
    """
    自动获取端口，返回数字端口号
    """
    if port == 'auto':
        # 获取所有被占用的端口号
        used_ports = get_used_ports(start=start, **kwargs)
        for i in range(start, 65535):
            if i not in used_ports:
                return int(i)
        raise ValueError('No available port')
    else:
        return port
    

def auto_worker_address(worker_address, host, port):
    import socket
    if worker_address != 'auto':
        return worker_address
    if host in ['localhost', '127.0.0.1']:
        return f'http://{host}:{port}'
    elif host == '0.0.0.0':
        ## TODO，此处需要改进，获取本机ip
        # 获取本机的外部 IP 地址是使用一个与外部世界的连接
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.connect(("10.255.255.255", 1))
            ip = s.getsockname()[0]
        return f'http://{ip}:{port}'
    else:
        raise ValueError(f'host {host} is not supported')


def extract_worker_id(text: str) -> str:
    import re
    # 使用正则表达式来匹配格式形如 'wk-xxxx-xxxx' 的 worker_id
    # match = re.search(r'worker_id: `(\w+-\w+)`', text)
    match = re.search(r'worker_id:\s*`([^`]*)`', text)
    if match:
        return match.group(1)
    return None

def wait_for_worker(worker_id, timeout=10):
    # 向服务发起请求
    import os, time
    from haiddf import HaiDDF
    from haiddf._types import WorkerInfo
    client = HaiDDF(api_key=os.getenv("DDF_API_KEY"))

    start_time = time.time()
    while time.time() - start_time < timeout:
        wk_info = client.get_worker_info(worker_id=worker_id)
        if isinstance(wk_info, WorkerInfo):
            return wk_info
        time.sleep(0.5)
    return False

def run_standlone_worker_demo(logger=None):
    import os, sys
    import time
    from pathlib import Path
    here = Path(__file__).parent

    class PrintLogger:
        def info(self, *args, **kwargs):
            print(*args, **kwargs)

    logger = logger or PrintLogger()
    
    current_dir = os.getcwd()

    import subprocess
    from datetime import datetime

    os.chdir(str(here))
    try:
        print(f"Run standalone worker demo in `{here}` ...", end="")
        # logger.info(f"Run standalone worker demo in `{here}` ...")
        ct = datetime.now().strftime("%Y%m%d_%H%M%S")
        log_file = f"nohup.out.demo.{ct}"
        cmd = f"nohup python demo_worker.py >{log_file} 2>&1 &"
        result = subprocess.run(
            cmd,
            shell=True,  
            check=True,  # 如果命令失败，抛出 CalledProcessError
            capture_output=True,  # 捕获标准输出和标准错误
            text=True  # 将输出作为字符串处理，而不是字节
        )
        print(", succeeded.", end="")
        # print(result.stdout)
    except subprocess.CalledProcessError as e:
        print(", failed.")
        print("Error message:")
        print(e.stderr)
        raise e

    os.chdir(current_dir)

    # 从logfile读取worker_id
    # 等待worker启动
    time.sleep(0.5)

    # 等待文件被创建和写入
    file_path = f"{here}/{log_file}"
    timeout = 5
    t0 = time.time() 
    while True:
        ct = time.time() - t0
        if ct > timeout:
            raise TimeoutError(f"Wait for worker start timeout. {ct}s")
        if not os.path.exists(file_path):
            time.sleep(0.5)
            # print(f"Waiting for file `{file_path}` to be created ... {ct}s")
            logger.info(f"Waiting for log_file `{log_file}` to be created ... {ct:.2f}s")
            continue
        
        with open(f'{here}/{log_file}', 'r') as f:
            text = f.read()
            if text == "":
                time.sleep(0.5)
                logger.info(f"Waiting for log_file `{log_file}` to be written ... {ct}s")
                # print(f"Waiting for file `{file_path}` to be written ... {ct}s")
                continue
        worker_id = extract_worker_id(text)
        # print(f" Worker_id: `{worker_id}`")
        if worker_id:
            return wait_for_worker(worker_id, timeout=10)
        else:
            if "Traceback" in text:
                chunks = text.split("Traceback")
                error_msg = f"Traceback{chunks[-1]}"
                raise ProcessLookupError(f"Worker started failed. {error_msg}")
            # print(" Worker started, but worker_id not found.")
            # return
            raise ValueError("Worker started, but worker_id not found.")
        




from hai import *

from hai import __version__
from hai import HepAI
from hai import HaiCompletions, ChatCompletion, ChatCompletionChunk, Stream
from hai import HaiFile


from .modules.haiddf.client.haiddf_client import HaiDDFClient



import os, sys
from pathlib import Path
from hai.version import __appname__

# /home/xxx/VSProjects/hai
ROOT_PATH = f'{Path(os.path.abspath(__file__)).parent.parent.parent.parent}'  

# /home/xxx/.hai/weights
WEIGHTS_ROOT = f'{Path.home()}/.{__appname__}/weights' 

# /home/xxx/.hai/datasets
DATASETS_ROOT = f'{Path.home()}/datasets/{__appname__}_datasets'

# hai_api
API_FOLD_NAME = f'{__appname__}_api' 




"""
基础类的定义
"""

import time
from dataclasses import dataclass, field, asdict
from typing import List, Dict, Literal, Union, Optional, Any


class BaseWorkerModel:

    @classmethod
    def remote_callable(cls, func):
        """
        用来修饰一个函数，使得其可以被远程调用，未修饰的函数不能被远程调用
        Decorator to mark a method as remotely callable.
        """
        func.is_remote_callable = True
        return func

class HWorkerModel(BaseWorkerModel):
    """
    The HepAI basic worker model class 
    """
    def __init__(self, name: str = "HWorkerModel"):
        self.name = name

    @BaseWorkerModel.remote_callable
    def hello_world(self, *args, **kwargs):
        """An example of a function that returns a string"""
        return f"Hello world! You are using the HepAI worker model with args: `{args}`, kwargs: `{kwargs}`"

    @BaseWorkerModel.remote_callable
    def get_int(self, a: int, b: int) -> int:
        """An example of a function that returns an int type"""
        return a + b
    
    @BaseWorkerModel.remote_callable
    def get_float(self, a: float, b: float) -> float:
        """An example of a function that returns a float type"""
        return a + b
    
    @BaseWorkerModel.remote_callable
    def get_list(self, a: List[int], b: List[int]) -> List[int]:
        """An example of a function that returns a list type"""
        return a + b
    
    @BaseWorkerModel.remote_callable
    def get_dict(self, a: Dict[str, int], b: Dict[str, int]) -> Dict[str, int]:
        """An example of a function that returns a dict type"""
        return {**a, **b}
    
    @BaseWorkerModel.remote_callable
    def get_stream(self, data: Any, interval: float = None):
        """An example of a function that returns a stream type"""

        for i, x in enumerate(data):
            time.sleep(interval)
            yield f"data: {x}\n\n"
            time.sleep(interval)

    @BaseWorkerModel.remote_callable
    def __call__(self, *args, **kwargs):
        return f"Hello world! You are calling function `__call__` of the HepAI remote model with args: `{args}`, kwargs: `{kwargs}`"


@dataclass
class ModelResourceInfo:
    """
    Model Resource Info for Worker, such as llm, nn, preceptor, actuator, etc.
    """
    model_name: str = field(default="<default_modelname>", metadata={"help": "Model's name"})
    model_type: str = field(default="common", metadata={"help": "Model's type"})
    model_version: str = field(default="1.0", metadata={"help": "Model's version"})
    model_description: str =field(default="<This is model description.>", metadata={"help": "Model's description"})
    model_author: Union[str, List[str], None] = field(default="", metadata={"help": "Model's author"})
    model_onwer: Union[str, None] = field(default="", metadata={"help": "Model's onwer"})
    model_groups: List[str] = field(default_factory=list, metadata={"help": "Model's groups"})
    model_users: List[str] = field(default_factory=list, metadata={"help": "Model's users"})
    model_functions: List[str] = field(default_factory=list, metadata={"help": "Model's functions that can be called by remote"})

@dataclass
class WorkerStatusInfo:
    """
    Worker Status Info, will be dynamic updated
    """
    speed: int = field(default=1, metadata={"help": "Worker's speed, the number of requests that can be processed per second"})
    queue_length: int = field(default=0, metadata={"help": "Worker's queue length"})
    status: Literal["idle", "ready", "busy", "error"] = "idle"

    def is_valid(self):
        """是信息是否可用，即相关信息是否已被填入，而不是None"""
        return self.speed > 0 and self.queue_length >= 0
    
    def to_dict(self):
        return asdict(self)


@dataclass
class WorkerNetworkInfo:
    """
    Network Info for Worker
    """
    host: str = field(default="127.0.0.1", metadata={"help": "Worker's host"})
    port: int = field(default=42602, metadata={"help": "Worker's port"})
    route_prefix: Union[str, None] = field(default="", metadata={"help": "Worker's route prefix, default is '/', the worker's address will be `http://host:port/route_prefix/other_router` if setted"})
    host_name: str = field(default="localhost", metadata={"help": "Worker's host name"})
    worker_address: Union[None, str] = field(default="", metadata={"help": "Worker's address, will be auto generated if not setted"})

    def check_and_autoset_worker_address(self):
        """自动检查并设置worker_address"""
        if self.worker_address in ["", None]:
            if self.route_prefix in ["", None]:
                self.worker_address = f"http://{self.host}:{self.port}"
            else:
                self.worker_address = f"http://{self.host}:{self.port}/{self.route_prefix}"

    def to_dict(self):
        return asdict(self)

@dataclass
class WorkerInfo:
    """
    Worker Info
    """
    id: str
    type: Literal["llm", "actuator", "preceptor", "memory", "common"] = "common"
    network_info: WorkerNetworkInfo = field(default_factory=WorkerNetworkInfo, metadata={"help": "Worker's network info"})
    resource_info: ModelResourceInfo = field(default_factory=ModelResourceInfo, metadata={"help": "Model's resource info"})
    status_info: WorkerStatusInfo = field(default_factory=WorkerStatusInfo, metadata={"help": "Worker's status info"})
    check_heartbeat: bool = True
    last_heartbeat: Union[int, None] = None
    vserion: str = "2.0"
    metadata: Dict = field(default_factory=dict, metadata={"help": "Worker's metadata"})

    def __post_init__(self):
        """在实例化本类时，自动检查network_info等是否Dict, 并转换为相应的对象"""
        if isinstance(self.network_info, dict):
            self.network_info = WorkerNetworkInfo(**self.network_info)
        if isinstance(self.resource_info, dict):
            self.resource_info = ModelResourceInfo(**self.resource_info)
        if isinstance(self.status_info, dict):
            self.status_info = WorkerStatusInfo(**self.status_info)

    def to_dict(self):
        return asdict(self)

from pydantic import BaseModel

class WorkerInfoItem(BaseModel):
    """用于FastAPI的WorkerInfo"""
    id: str = field(default="wk-default_id", metadata={"help": "Worker's id"})     
    type: Literal["llm", "actuator", "preceptor", "memory", "common"] = "common"
    network_info: WorkerNetworkInfo = field(default_factory=WorkerNetworkInfo, metadata={"help": "Worker's network info"})
    resource_info: ModelResourceInfo = field(default_factory=ModelResourceInfo, metadata={"help": "Model's resource info"})
    status_info: WorkerStatusInfo = field(default_factory=WorkerStatusInfo, metadata={"help": "Worker's status info"})
    check_heartbeat: bool = field(default=True, metadata={"help": "Check worker's heartbeat"})
    last_heartbeat: Union[int, None] = field(default=None, metadata={"help": "Worker's last heartbeat"})
    vserion: str = field(default="2.0", metadata={"help": "Worker's version"})
    metadata: Dict = field(default_factory=dict, metadata={"help": "Worker's metadata"})


class WorkerStoppedInfo(BaseModel):
    id: str = field(metadata={"help": "Worker's id"}) 
    stopped: bool = field(metadata={"help": "Worker's stopped flag"})
    message: str = field(default=None, metadata={"help": "Worker's stopped message"})
    shutdown: bool = field(default=False, metadata={"help": "Worker's shutdown"})

    def to_dict(self):
        return {
            "id": self.id,
            "stopped": self.stopped,
            "message": self.message,
            "shutdown": self.shutdown,
        }




def pretty_print_semaphore(semaphore):
    if semaphore is None:
        return "None"
    return f"Semaphore(value={semaphore._value}, locked={semaphore.locked()})"

def get_used_ports(start=1, end=65535):
    import socket
    ports = []
    for port in range(start, end):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind(('0.0.0.0', port))
        except:
            ports.append(port)
        finally:
            s.close()
    return ports

def auto_port(port=None, start=42901, **kwargs):
    """
    自动获取端口，返回数字端口号
    """
    if port == 'auto':
        # 获取所有被占用的端口号
        used_ports = get_used_ports(start=start, **kwargs)
        for i in range(start, 65535):
            if i not in used_ports:
                return int(i)
        raise ValueError('No available port')
    else:
        return port
    
def auto_worker_address(worker_address, host, port):
    import socket
    if worker_address != 'auto':
        return worker_address
    if host in ['localhost', '127.0.0.1']:
        return f'http://{host}:{port}'
    elif host == '0.0.0.0':
        ## TODO，此处需要改进，获取本机ip
        # 获取本机的外部 IP 地址是使用一个与外部世界的连接
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.connect(("10.255.255.255", 1))
            ip = s.getsockname()[0]
        return f'http://{ip}:{port}'
    else:
        raise ValueError(f'host {host} is not supported')


    





from .modules.haiddf.base_class._user_class import *
from .modules.haiddf.base_class._worker_class import *
from .modules.haiddf.client._client_class import *
from .modules.haiddf.client.resources.worker_resource import HRemoteModel

from .modules.haiddf.worker._worker_class import *
from .modules.haiddf.worker.worker_app import HWorkerAPP